"""
Pandas accessors for OpenSCM-related functionality.
"""

import importlib.metadata

from pandas_openscm.accessors import register_pandas_accessor

__version__ = importlib.metadata.version("pandas_openscm")

__all__ = ["register_pandas_accessor"]



"""
Tests of saving and loading with `pandas_openscm.OpenSCMDB`

Note that these are also supplemented by our state testing with hypothesis
(`tests/integration/database/test_integration_database_state.py`),
hence we don't have to test every combination here.
"""

from __future__ import annotations

import re
from contextlib import nullcontext
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    AlreadyInDBError,
    InMemoryDataBackend,
    InMemoryIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.index_manipulation import unify_index_levels
from pandas_openscm.testing import (
    assert_frame_alike,
    create_test_df,
    get_parametrized_db_data_backends,
    get_parametrized_db_index_backends,
)


@get_parametrized_db_data_backends()
@get_parametrized_db_index_backends()
def test_save_and_load_basic(tmpdir, db_data_backend, db_index_backend):
    if "Feather" in str(db_data_backend) or "Feather" in str(db_index_backend):
        pytest.importorskip("pyarrow")

    if "netCDF" in str(db_data_backend) or "netCDF" in str(db_index_backend):
        pytest.importorskip("xarray")

    df_timeseries_like = pd.DataFrame(
        np.arange(12).reshape(4, 3),
        columns=[2010, 2015, 2025],
        index=pd.MultiIndex.from_tuples(
            [
                ("scenario_a", "climate_model_a", "Temperature", "K"),
                ("scenario_b", "climate_model_a", "Temperature", "K"),
                ("scenario_b", "climate_model_b", "Temperature", "K"),
                ("scenario_b", "climate_model_b", "Ocean Heat Uptake", "J"),
            ],
            names=["scenario", "climate_model", "variable", "unit"],
        ),
    )

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=db_data_backend(),
        backend_index=db_index_backend(),
        index_file_lock=nullcontext(),  # not used
    )

    db.save(df_timeseries_like)

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(df_timeseries_like.index.names)
    pd.testing.assert_index_equal(
        df_timeseries_like.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=df_timeseries_like.columns.dtype)

    assert_frame_alike(df_timeseries_like, loaded)

    # Check the file map, index and metadata too
    file_map = db.load_file_map()
    assert file_map.size == 1, "should only be one file"

    index = db.load_index()
    assert index.columns.tolist() == ["file_id"]
    pd.testing.assert_index_equal(
        index.index,
        df_timeseries_like.index.reorder_levels(index.index.names),
        check_order=False,
    )

    metadata = db.load_metadata()
    pd.testing.assert_index_equal(
        metadata,
        df_timeseries_like.index.reorder_levels(metadata.names),
        check_order=False,
    )


def test_save_and_load(tmpdir):
    start = create_test_df(
        n_scenarios=10,
        variables=[("a", "kg"), ("b", "kg"), ("c", "kg")],
        n_runs=5,
        timepoints=np.arange(2015.0, 2100.0, 5.0),
    )

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    db.save(start)

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(start.index.names)
    pd.testing.assert_index_equal(
        start.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=start.columns.dtype)

    assert_frame_alike(start, loaded)

    # Check the file map, index and metadata too
    file_map = db.load_file_map()
    assert file_map.size == 1, "should only be one file"

    index = db.load_index()
    assert index.columns.tolist() == ["file_id"]
    pd.testing.assert_index_equal(
        index.index, start.index.reorder_levels(index.index.names), check_order=False
    )

    metadata = db.load_metadata()
    pd.testing.assert_index_equal(
        metadata, start.index.reorder_levels(metadata.names), check_order=False
    )


def test_save_multiple_and_load(tmpdir):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    all_saved_l = []
    for variable in [
        [("Emissions", "Gt")],
        [("Concentrations", "ppm")],
        [("Forcing", "W/m^2")],
    ]:
        to_save = create_test_df(
            n_scenarios=10,
            variables=variable,
            n_runs=3,
            timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
        )

        db.save(to_save)
        all_saved_l.append(to_save)

    all_saved = pix.concat(all_saved_l)

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(all_saved.index.names)
    pd.testing.assert_index_equal(
        all_saved.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(all_saved, loaded)


def test_save_multiple_grouped_and_load(tmpdir):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    all_saved_l = []
    for variable in [
        [("Emissions", "Gt")],
        [("Concentrations", "ppm")],
        [("Forcing", "W/m^2")],
    ]:
        to_save = create_test_df(
            n_scenarios=10,
            variables=variable,
            n_runs=3,
            timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
        )

        db.save(to_save, groupby=["scenario", "variable"])
        all_saved_l.append(to_save)

    all_saved = pix.concat(all_saved_l)

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(all_saved.index.names)
    pd.testing.assert_index_equal(
        all_saved.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(all_saved, loaded)


@pytest.mark.parametrize(
    "wide_first",
    (
        pytest.param(True, id="wide-first"),
        pytest.param(False, id="narrow-first"),
    ),
)
def test_save_multiple_grouped_wide_and_narrow_and_load(wide_first, tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    to_save_wide_index = create_test_df(
        n_scenarios=3,
        variables=[("Emission", "Gt"), ("Concentrations", "ppm")],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    to_save_narrow_index = to_save_wide_index.groupby(
        ["variable", "unit", "run"]
    ).mean()
    assert len(to_save_narrow_index.index.names) < len(to_save_wide_index.index.names)

    if wide_first:
        db.save(to_save_wide_index.copy(), groupby=["scenario", "variable"])
        db.save(to_save_narrow_index.copy(), groupby=["variable", "unit"])
    else:
        db.save(to_save_narrow_index.copy(), groupby=["variable", "unit"])
        db.save(to_save_wide_index.copy(), groupby=["scenario", "variable"])

    tmp = unify_index_levels(to_save_wide_index.index, to_save_narrow_index.index)[1]
    to_save_narrow_index_unified_index = to_save_narrow_index.copy()
    to_save_narrow_index_unified_index.index = tmp
    all_saved_exp = pd.concat([to_save_wide_index, to_save_narrow_index_unified_index])

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(all_saved_exp.index.names)
    pd.testing.assert_index_equal(
        all_saved_exp.index, metadata_compare, check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(all_saved_exp, loaded)


def test_save_overwrite_error(tmpdir):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    cdf = partial(
        create_test_df,
        n_scenarios=10,
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    dup = cdf(variables=[("Emissions", "t")])
    db.save(dup)

    to_save = pix.concat([dup, cdf(variables=[("Weight", "kg")])])

    error_msg = re.escape(
        "The following rows are already in the database:\n"
        f"{dup.index.to_frame(index=False)}"
    )
    with pytest.raises(AlreadyInDBError, match=error_msg):
        db.save(to_save)


def test_save_overwrite_force(tmpdir):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    cdf = partial(
        create_test_df,
        n_scenarios=10,
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    original = cdf(variables=[("Emissions", "t")])
    db.save(original)

    # Make sure that our data saved correctly
    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(original.index.names)
    pd.testing.assert_index_equal(
        original.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(original, loaded)

    original_overwrite = cdf(variables=[("Emissions", "t")])
    updated = pix.concat([original_overwrite, cdf(variables=[("Height", "m")])])

    # With force, we can overwrite
    db.save(updated, allow_overwrite=True)

    # Check that the data was overwritten with new data
    try:
        assert_frame_alike(original, original_overwrite)
    except AssertionError:
        pass
    else:
        # Somehow got the same DataFrame,
        # so there won't be any difference in the db.
        msg = "Test won't do anything"
        raise AssertionError(msg)

    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(updated.index.names)
    pd.testing.assert_index_equal(
        updated.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(updated, loaded)


@pytest.mark.parametrize(
    "warn_on_partial_overwrite,expectation_overwrite_warning",
    (
        pytest.param(
            None,
            pytest.warns(
                match="Overwriting the data will require re-writing. This may be slow."
            ),
            id="default",
        ),
        pytest.param(
            True,
            pytest.warns(
                match="Overwriting the data will require re-writing. This may be slow."
            ),
            id="explicitly-enabled",
        ),
        pytest.param(False, nullcontext(), id="silenced"),
    ),
)
def test_save_overwrite_force_half_overlap(
    warn_on_partial_overwrite,
    expectation_overwrite_warning,
    tmpdir,
):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    cdf = partial(
        create_test_df,
        variables=[(f"v_{i}", "m") for i in range(5)],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    original = cdf(n_scenarios=6)
    db.save(original)

    original_overwrite = cdf(n_scenarios=3)

    # Check that the data was overwritten with new data
    overlap_idx = original.index.isin(original_overwrite.index)
    overlap = original.loc[overlap_idx]
    try:
        assert_frame_alike(overlap, original_overwrite)
    except AssertionError:
        pass
    else:
        # Somehow got the same values,
        # so there won't be any difference in the db.
        msg = "Test won't do anything"
        raise AssertionError(msg)

    call_kwargs = {}
    if warn_on_partial_overwrite is not None:
        call_kwargs["warn_on_partial_overwrite"] = warn_on_partial_overwrite

    with expectation_overwrite_warning:
        db.save(original_overwrite, allow_overwrite=True, **call_kwargs)

    update_exp = pix.concat([original.loc[~overlap_idx], original_overwrite])
    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(update_exp.index.names)
    pd.testing.assert_index_equal(
        update_exp.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(update_exp, loaded)


# Parameterise here because this is so fundamental.
# For rest of checking backend combos,
# rely on state tests and design of the codebase
# (where backend isn't involved in determining groupings for data).
@get_parametrized_db_data_backends()
@get_parametrized_db_index_backends()
def test_save_overwrite_force_half_overlap_all_backends(
    db_data_backend,
    db_index_backend,
    tmpdir,
):
    pix = pytest.importorskip("pandas_indexing")

    if "Feather" in str(db_data_backend) or "Feather" in str(db_index_backend):
        pytest.importorskip("pyarrow")

    if "netCDF" in str(db_data_backend) or "netCDF" in str(db_index_backend):
        pytest.importorskip("xarray")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=db_data_backend(),
        backend_index=db_index_backend(),
        index_file_lock=nullcontext(),  # not used
    )

    cdf = partial(
        create_test_df,
        variables=[(f"v_{i}", "m") for i in range(5)],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    original = cdf(n_scenarios=6)
    db.save(original)

    # Make sure that our data saved correctly
    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(original.index.names)
    pd.testing.assert_index_equal(
        original.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(original, loaded)

    original_overwrite = cdf(n_scenarios=3)

    # Check that the data was overwritten with new data
    overlap_idx = original.index.isin(original_overwrite.index)
    overlap = original.loc[overlap_idx]
    try:
        assert_frame_alike(overlap, original_overwrite)
    except AssertionError:
        pass
    else:
        # Somehow got the same values,
        # so there won't be any difference in the db.
        msg = "Test won't do anything"
        raise AssertionError(msg)

    db.save(original_overwrite, allow_overwrite=True, warn_on_partial_overwrite=False)

    update_exp = pix.concat([original.loc[~overlap_idx], original_overwrite])
    db_metadata = db.load_metadata()
    metadata_compare = db_metadata.reorder_levels(update_exp.index.names)
    pd.testing.assert_index_equal(
        update_exp.index, metadata_compare, exact="equiv", check_order=False
    )

    loaded = db.load(out_columns_type=float)

    assert_frame_alike(update_exp, loaded)


def test_load_with_loc(tmpdir):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    full_db = create_test_df(
        n_scenarios=10,
        variables=[("variable_1", "kg"), ("variable_2", "Mt"), ("variable_3", "m")],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    db.save(full_db, groupby=["scenario"])

    for selector in [
        pix.isin(scenario=["scenario_1", "scenario_3"]),
        pix.isin(scenario=["scenario_1", "scenario_3"], variable=["variable_2"]),
        (
            pix.isin(scenario=["scenario_1", "scenario_3"])
            & pix.ismatch(variable=["variable_1*"])
        ),
        pix.isin(scenario=["scenario_1", "scenario_3"], variable=["variable_2"]),
    ]:
        loaded = db.load(selector, out_columns_type=float)
        exp = full_db.loc[selector]

        assert_frame_alike(loaded, exp)


def test_load_with_index_all(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    full_db = create_test_df(
        n_scenarios=10,
        variables=[("variable_1", "kg"), ("variable_2", "Mt"), ("variable_3", "m")],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    db.save(full_db, groupby=["scenario"])

    idx = full_db.index
    exp = full_db

    loaded = db.load(idx, out_columns_type=float)

    assert_frame_alike(loaded, exp)


@pytest.mark.parametrize(
    "slice",
    (slice(None, None, None), slice(None, 3, None), slice(2, 4, None), slice(1, 15, 2)),
)
def test_load_with_index_slice(tmpdir, slice):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    full_db = create_test_df(
        n_scenarios=10,
        variables=[("variable_1", "kg"), ("variable_2", "Mt"), ("variable_3", "m")],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    db.save(full_db, groupby=["scenario"])

    idx = full_db.index[slice]
    exp = full_db[slice]

    loaded = db.load(idx, out_columns_type=float)

    assert_frame_alike(loaded, exp)


@pytest.mark.parametrize(
    "levels",
    (
        pytest.param(["scenario"], id="first_level"),
        pytest.param(["variable"], id="not_first_level"),
        pytest.param(["scenario", "variable"], id="multi_level_in_order"),
        pytest.param(["scenario", "variable"], id="multi_level_non_adjacent"),
        pytest.param(["variable", "scenario"], id="multi_level_out_of_order"),
        pytest.param(["run", "variable"], id="multi_level_out_of_order_not_first"),
    ),
)
def test_load_with_pix_unique_levels(tmpdir, levels):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    full_db = create_test_df(
        n_scenarios=10,
        variables=[("variable_1", "kg"), ("variable_2", "Mt"), ("variable_3", "m")],
        n_runs=3,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    db.save(full_db, groupby=["scenario"])

    locator = None
    for level in levels:
        if locator is None:
            locator = pix.isin(**{level: full_db.pix.unique(level)[:2]})
        else:
            locator &= pix.isin(**{level: full_db.pix.unique(level)[:2]})

    exp = full_db.loc[locator]
    idx = exp.pix.unique(levels)

    loaded = db.load(idx, out_columns_type=float)

    assert_frame_alike(loaded, exp)



"""
[Feather](https://arrow.apache.org/docs/python/feather.html) backend
"""

from __future__ import annotations

from pathlib import Path
from typing import Literal

import pandas as pd
from attrs import define


@define
class FeatherDataBackend:
    """
    Feather data backend

    For details on feather, see https://arrow.apache.org/docs/python/feather.html
    """

    ext: str = ".feather"
    """
    Extension to use with files saved by this backend.
    """

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the index of data upon (de-)serialisation
        """
        return True

    @staticmethod
    def load_data(data_file: Path) -> pd.DataFrame:
        """
        Load a data file

        Parameters
        ----------
        data_file
            File from which to load the data

        Returns
        -------
        :
            Loaded data
        """
        return pd.read_feather(data_file)

    @staticmethod
    def save_data(data: pd.DataFrame, data_file: Path) -> None:
        """
        Save data to disk

        Parameters
        ----------
        data
            Data to save

        data_file
            File in which to save the data
        """
        # The docs say that feather doesn't support writing indexes
        # # (see https://pandas.pydata.org/docs/user_guide/io.html#feather).
        # However, it seems to have no issue writing our multi-indexes.
        # Hence the implementation below
        data.to_feather(data_file)


@define
class FeatherIndexBackend:
    """
    Feather index backend

    For details on feather, see https://arrow.apache.org/docs/python/feather.html
    """

    ext: str = ".feather"
    """
    Extension to use with files saved by this backend.
    """

    @property
    def preserves_index(self) -> Literal[True]:
        """
        Whether this backend preserves the `pd.MultiIndex` upon (de-)serialisation
        """
        return True

    @staticmethod
    def load_file_map(file_map_file: Path) -> pd.DataFrame:
        """
        Load the file map

        Parameters
        ----------
        file_map_file
            File from which to load the file map

        Returns
        -------
        :
            Loaded file map
        """
        return pd.read_feather(file_map_file)

    @staticmethod
    def load_index(index_file: Path) -> pd.DataFrame:
        """
        Load the index

        Parameters
        ----------
        index_file
            File from which to load the index

        Returns
        -------
        :
            Loaded index
        """
        return pd.read_feather(index_file)

    @staticmethod
    def save_file_map(
        file_map: pd.Series[Path],  # type: ignore # pandas confused about what it supports
        file_map_file: Path,
    ) -> None:
        """
        Save the file map to disk

        Parameters
        ----------
        file_map
            File map to save

        file_map_file
            File in which to save the file map
        """
        # Feather doesn't support writing non-native types
        # (see https://pandas.pydata.org/docs/user_guide/io.html#feather).
        # The docs say that feather doesn't support writing indexes
        # # (see https://pandas.pydata.org/docs/user_guide/io.html#feather).
        # However, it seems to have no issue writing this index.
        # Hence the implementation below
        file_map_write = file_map.astype(str)
        file_map_write.to_frame().to_feather(file_map_file)

    @staticmethod
    def save_index(
        index: pd.DataFrame,
        index_file: Path,
    ) -> None:
        """
        Save the index to disk

        Parameters
        ----------
        index
            Index to save

        index_file
            File in which to save the index
        """
        index.to_feather(index_file)



"""
Tests of reading using an in-memory index with `pandas_openscm`
"""

from __future__ import annotations

import contextlib
from contextlib import nullcontext as does_not_raise
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    CSVDataBackend,
    CSVIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.testing import assert_frame_alike, create_test_df


def check_metadata_load_is_same(
    reader_metadata: pd.MultiIndex, db_metadata: pd.MultiIndex
) -> None:
    metadata_compare = reader_metadata.reorder_levels(db_metadata.names)
    pd.testing.assert_index_equal(
        db_metadata, metadata_compare, exact="equiv", check_order=False
    )


def test_load_via_reader_context_manager(tmpdir):
    pytest.importorskip("filelock")

    start = create_test_df(
        n_scenarios=10,
        variables=[("a", "kg"), ("b", "kg"), ("c", "kg")],
        n_runs=5,
        timepoints=np.arange(2015.0, 2100.0, 5.0),
    )

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=CSVDataBackend(),
        backend_index=CSVIndexBackend(),
    )

    db.save(start)

    db_metadata = db.load_metadata()

    with db.create_reader() as reader:
        assert reader.lock.is_locked

        check_metadata_load_is_same(reader.metadata, db_metadata)

        loaded = reader.load(out_columns_type=start.columns.dtype)
        assert_frame_alike(start, loaded)

    # Lock released on exiting the context block
    assert not reader.lock.is_locked


def test_load_via_reader(tmpdir):
    start = create_test_df(
        n_scenarios=10,
        variables=[("a", "kg"), ("b", "kg"), ("c", "kg")],
        n_runs=5,
        timepoints=np.arange(2015.0, 2100.0, 5.0),
    )

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=CSVDataBackend(),
        backend_index=CSVIndexBackend(),
        index_file_lock=contextlib.nullcontext(),  # not used
    )

    db.save(start)

    reader = db.create_reader(lock=False)

    db_metadata = db.load_metadata()
    check_metadata_load_is_same(reader.metadata, db_metadata)

    loaded = reader.load(out_columns_type=start.columns.dtype)

    assert_frame_alike(start, loaded)


def test_reader_locking(tmpdir):
    """
    Test the handling of locking via the reader

    This could be split out to provide better diagnostic power of issues.

    However, this overall flow is also a handy integration test
    so I would suggest keeping this test as is
    and just adding more tests
    (at least until maintaining this test becomes annoying).
    """
    filelock = pytest.importorskip("filelock")

    start = create_test_df(
        n_scenarios=10,
        variables=[("a", "kg"), ("b", "kg"), ("c", "kg")],
        n_runs=5,
        timepoints=np.arange(2015.0, 2100.0, 5.0),
    )

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=CSVDataBackend(),
        backend_index=CSVIndexBackend(),
    )

    db.save(start)

    # If we use the context manager,
    # the created object holds the lock
    # so we can't read/write with the db anymore.
    with db.create_reader() as reader:
        # The db and the reader get different locks by default
        assert db.index_file_lock != reader.lock
        # The db's lock isn't locked
        assert not db.index_file_lock.is_locked
        # The reader's lock is locked
        assert reader.lock.is_locked

        with pytest.raises(filelock.Timeout):
            # We can't acquire the db's lock,
            # because the reader has the lock.
            db.index_file_lock.acquire(timeout=0.02)

    # Once we're out of the context block, the lock is released
    with does_not_raise():
        db.index_file_lock.acquire(timeout=0.02)

    db.index_file_lock.release()

    # You can bypass holding the lock within the context manager
    # (doesn't make much sense, just don't use the context manager in this case,
    # but at least it doesn't explode).
    with db.create_reader(lock=False) as reader:
        assert reader.lock is None

        with does_not_raise():
            db.index_file_lock.acquire(timeout=0.02)

        db.index_file_lock.release()

    # By default, using `create_reader` does not hold the lock.
    reader = db.create_reader()

    # So we could keep doing database ops
    db.index_file_lock.acquire()
    # (Release again to avoid double releasing later)
    db.index_file_lock.release()

    # However, if we now acquire the reader's lock
    reader.lock.acquire()
    # The db is locked again
    with pytest.raises(filelock.Timeout):
        db.index_file_lock.acquire(timeout=0.02)

    # If we release the lock.
    reader.lock.release()

    # We can get the lock again
    with does_not_raise():
        db.index_file_lock.acquire(timeout=0.02)

    # This is a no op,
    # but this checks that calling again doesn't cause anything to explode
    reader.lock.release()

    # We can also create the reader within a context block that is managing the lock.
    # We do have to be a bit more careful in this case.
    with db.index_file_lock as lock:
        # By default, everything locks
        # because the reader is trying to acquire and it can't.
        # with db.create_reader() as reader:
        #     ...
        # This code illustrates the above without making the test hang forever.
        with pytest.raises(filelock.Timeout):
            with db.create_reader(
                lock=filelock.FileLock(db.index_file_lock_path, timeout=0.02)
            ) as reader:
                ...

        # If we pass in the lock we're already managing, then there is no issue
        # and we can get information from either path.
        with db.create_reader(lock=lock) as reader:
            check_metadata_load_is_same(reader.metadata, db.load_metadata())
            assert_frame_alike(reader.load(), db.load())



"""
Basic unit tests of `pandas_openscm.database`
"""

from __future__ import annotations

import re
import sys
from contextlib import nullcontext
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    DATA_BACKENDS,
    INDEX_BACKENDS,
    CSVDataBackend,
    CSVIndexBackend,
    EmptyDBError,
    FeatherDataBackend,
    FeatherIndexBackend,
    InMemoryDataBackend,
    InMemoryIndexBackend,
    OpenSCMDB,
    netCDFDataBackend,
    netCDFIndexBackend,
)
from pandas_openscm.exceptions import MissingOptionalDependencyError
from pandas_openscm.testing import create_test_df


def test_available_backends_data():
    assert isinstance(DATA_BACKENDS.get_instance("csv"), CSVDataBackend)
    assert isinstance(DATA_BACKENDS.get_instance("feather"), FeatherDataBackend)
    assert isinstance(DATA_BACKENDS.get_instance("in_memory"), InMemoryDataBackend)
    assert isinstance(DATA_BACKENDS.get_instance("netCDF"), netCDFDataBackend)


def test_unavailable_data_backend():
    with pytest.raises(KeyError):
        DATA_BACKENDS.get_instance("junk")


def test_available_backends_index():
    assert isinstance(INDEX_BACKENDS.get_instance("csv"), CSVIndexBackend)
    assert isinstance(INDEX_BACKENDS.get_instance("feather"), FeatherIndexBackend)
    assert isinstance(INDEX_BACKENDS.get_instance("in_memory"), InMemoryIndexBackend)
    assert isinstance(INDEX_BACKENDS.get_instance("netCDF"), netCDFIndexBackend)


def test_unavailable_index_backend():
    with pytest.raises(KeyError):
        INDEX_BACKENDS.get_instance("junk")


def test_filelock_not_available_default_initialisation(tmpdir):
    with patch.dict(sys.modules, {"filelock": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=("`default_index_file_lock` requires filelock to be installed"),
        ):
            OpenSCMDB(
                db_dir=Path(tmpdir),
                backend_data=InMemoryDataBackend(),
                backend_index=InMemoryIndexBackend(),
            )

        # Not an issue if we bypass the lock
        OpenSCMDB(
            db_dir=Path(tmpdir),
            backend_data=InMemoryDataBackend(),
            backend_index=InMemoryIndexBackend(),
            index_file_lock=nullcontext(),
        )


def test_get_existing_data_file_path(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    fp = db.get_new_data_file_path(file_id=10)

    # Assume the file gets written somewhere
    fp.touch()

    with pytest.raises(FileExistsError):
        db.get_new_data_file_path(file_id=10)


@pytest.mark.parametrize(
    "meth, args, expecation",
    (
        *[
            (meth, args, nullcontext())
            for meth, args in [
                ("delete", []),
                ("get_new_data_file_path", [0]),
                (
                    "save",
                    [
                        create_test_df(
                            variables=(("variable", "kg"),),
                            n_scenarios=1,
                            n_runs=1,
                            timepoints=np.array([1.0, 1.5]),
                        )
                    ],
                ),
            ]
        ],
        *[
            (
                meth,
                args,
                pytest.raises(EmptyDBError, match="The database is empty: db="),
            )
            for meth, args in [
                ("load", []),
                ("load_file_map", []),
                ("load_index", []),
                ("load_metadata", []),
            ]
        ],
    ),
)
def test_raise_if_empty(tmpdir, meth, args, expecation):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    with expecation:
        getattr(db, meth)(*args)


def test_save_data_index_not_multi_error(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    data = pd.DataFrame([0, 1], index=pd.Index(["a", "b"]))
    with pytest.raises(
        TypeError,
        match=re.escape(
            "`data.index` must be an instance of `pd.MultiIndex`. "
            "Received type(data.index)=<class 'pandas.core.indexes.base.Index'>"
        ),
    ):
        db.save(data)


def test_save_data_duplicate_index_rows(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),  # not used
    )

    data = pd.DataFrame(
        [
            [1.0, 2.0],
            [8.0, 1.0],
            [9.0, 7.0],
            [3.0, 5.0],
        ],
        columns=[2010.0, 2020.0],
        index=pd.MultiIndex.from_tuples(
            [
                ("scen_a", "mod_a", "var_a"),
                ("scen_b", "mod_b", "var_b"),
                ("scen_a", "mod_a", "var_a"),
                ("scen_a", "mod_b", "var_b"),
            ],
            names=["scenario", "model", "variable"],
        ),
    )

    duplicates = data.loc[data.index.duplicated(keep=False)]
    with pytest.raises(
        ValueError,
        match=re.escape(
            f"`data` contains rows with the same metadata. duplicates=\n{duplicates}"
        ),
    ):
        db.save(data)


def test_filelock_not_available_default_reader_lock(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=nullcontext(),
    )
    # Need some data in there so the reader can be created
    data = pd.DataFrame(
        [[1.0, 2.0]],
        columns=[2010.0, 2020.0],
        index=pd.MultiIndex.from_tuples(
            [("scen_a", "mod_a", "var_a")],
            names=["scenario", "model", "variable"],
        ),
    )
    db.save(data)

    with patch.dict(sys.modules, {"filelock": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`create_reader(..., lock=True, ...)` requires filelock to be installed"
            ),
        ):
            db.create_reader()

        # If we disable lock creation, all fine
        db.create_reader(lock=False)

        # Or pass in our own lock, also all fine
        db.create_reader(lock=nullcontext())



"""
Tests of `pandas_openscm.plotting` and `pd.DataFrame.openscm.plot*`
"""

from __future__ import annotations

import contextlib
import re
import sys
from contextlib import nullcontext as does_not_raise
from pathlib import Path
from typing import Any
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.exceptions import MissingOptionalDependencyError
from pandas_openscm.plotting import (
    PlumePlotter,
    SingleLinePlotter,
    extract_single_unit,
    get_default_colour_cycler,
    get_quantiles,
    get_values_line,
    get_values_plume,
    plot_plume_after_calculating_quantiles_func,
    plot_plume_func,
)
from pandas_openscm.testing import create_test_df

matplotlib = pytest.importorskip("matplotlib")
plt = pytest.importorskip("matplotlib.pyplot")
pytest_regressions = pytest.importorskip("pytest_regressions")

try:
    import openscm_units
except ImportError:
    openscm_units = None


def check_plots(
    plot_kwargs: dict[str, Any],
    df: pd.DataFrame,
    image_regression: pytest_regressions.image_regression.ImageRegressionFixture,
    tmp_path: Path,
    exp: contextlib.AbstractContextManager = does_not_raise(),
) -> None:
    fig, ax = plt.subplots()

    with exp:
        return_val = plot_plume_func(df, ax=ax, **plot_kwargs)

    assert return_val == ax

    out_file = tmp_path / "fig.png"
    plt.savefig(out_file, bbox_extra_artists=(ax.get_legend(),), bbox_inches="tight")

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    # Check this works via the accessor too
    fig, ax = plt.subplots()
    with exp:
        return_val = df.openscm.plot_plume(ax=ax, **plot_kwargs)

    assert return_val == ax

    out_file = tmp_path / "fig-accessor.png"
    plt.savefig(out_file, bbox_extra_artists=(ax.get_legend(),), bbox_inches="tight")

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    plt.close()


def check_plots_incl_quantile_calculation(
    method_kwargs: dict[str, Any],
    df: pd.DataFrame,
    image_regression: pytest_regressions.image_regression.ImageRegressionFixture,
    tmp_path: Path,
    exp: contextlib.AbstractContextManager = does_not_raise(),
) -> None:
    fig, ax = plt.subplots()

    with exp:
        return_val = plot_plume_after_calculating_quantiles_func(
            df, ax=ax, **method_kwargs
        )

    assert return_val == ax

    out_file = tmp_path / "fig.png"
    plt.savefig(out_file, bbox_extra_artists=(ax.get_legend(),), bbox_inches="tight")

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    # Check this works via the accessor too
    fig, ax = plt.subplots()
    with exp:
        return_val = df.openscm.plot_plume_after_calculating_quantiles(
            ax=ax, **method_kwargs
        )

    assert return_val == ax

    out_file = tmp_path / "fig-accessor.png"
    plt.savefig(out_file, bbox_extra_artists=(ax.get_legend(),), bbox_inches="tight")

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    plt.close()


def test_plot_plume_default(tmp_path, image_regression, setup_pandas_accessor):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=82747),
    )

    check_plots(
        df=df.openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile(),
        plot_kwargs=dict(quantiles_plumes=((0.5, 0.8), ((0.05, 0.95), 0.3))),
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


def test_default_ax_auto_creation(tmp_path, image_regression, setup_pandas_accessor):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=83747),
    )

    res = (
        df.openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile()
        .openscm.plot_plume(quantiles_plumes=((0.5, 0.8), ((0.05, 0.95), 0.3)))
    )
    assert isinstance(res, matplotlib.axes.Axes)

    out_file = tmp_path / "fig.png"
    plt.savefig(out_file, bbox_extra_artists=(res.get_legend(),), bbox_inches="tight")

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)


def test_plot_plume_no_labels(tmp_path, image_regression, setup_pandas_accessor):
    df = create_test_df(
        variables=(("variable_1", "K"),),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=84747),
    )

    check_plots(
        df=df.openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile(),
        plot_kwargs=dict(
            y_label=None,
            x_label=None,
            quantiles_plumes=((0.5, 0.8), ((0.05, 0.95), 0.3)),
        ),
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


def test_plot_plume_with_other_plot_calls(
    tmp_path, image_regression, setup_pandas_accessor
):
    fig, ax = plt.subplots()

    before_handles = ax.plot(
        np.arange(1955.0, 1975.0, 2.5),
        8 - np.arange(8),
        color="gray",
        label="before",
        linewidth=2.5,
        linestyle="--",
        zorder=3.1,
    )

    df = (
        create_test_df(
            variables=(("variable_1", "K"), ("variable_2", "K")),
            n_scenarios=3,
            n_runs=10,
            timepoints=np.arange(1950.0, 1965.0),
            rng=np.random.default_rng(seed=999),
        )
        .openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile()
    )

    df.openscm.plot_plume(
        ax=ax,
        quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
    )

    x_after = np.linspace(1940.0, 1970.0, 101)
    after_handles = ax.plot(
        x_after,
        np.sin(x_after) + 5.0 + np.linspace(0, 30.0, x_after.size),
        color="pink",
        label="after",
        linewidth=3,
        zorder=3.1,
    )

    # This is mucking around.
    # If you wanted to do something like this,
    # much simpler to just get the PlumePlotter object
    # and do the handles that way.
    # This at least shows that this sort of thing is possible.
    new_handles = [
        *before_handles,
        *after_handles,
        *ax.get_legend().legend_handles,
    ]
    ax.legend(handles=new_handles, loc="lower left", bbox_to_anchor=(1.05, 0.25))

    out_file = tmp_path / "fig.png"
    plt.savefig(
        out_file,
        bbox_extra_artists=(ax.get_legend(),),
        bbox_inches="tight",
    )

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    plt.close()


@pytest.mark.parametrize(
    "quantiles_plumes",
    (
        pytest.param(
            (
                (0.5, 0.7),
                ((0.25, 0.75), 0.5),
                ((0.05, 0.95), 0.2),
            ),
            id="multi-plume",
        ),
        pytest.param(
            (
                ((0.25, 0.75), 0.5),
                ((0.05, 0.95), 0.2),
            ),
            id="plumes-only",
        ),
        pytest.param(
            ((0.5, 0.7),),
            id="line-only",
        ),
        # If you actually wanted to do this,
        # you would just use the seaborn API directly,
        # but this at least checks that things don't explode.
        pytest.param(
            (
                (0.5, 0.7),
                (0.05, 0.7),
                (0.95, 0.7),
            ),
            id="lines-only",
        ),
    ),
)
def test_plot_plume_quantiles(
    quantiles_plumes, tmp_path, image_regression, setup_pandas_accessor
):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=11241),
    )

    plot_kwargs = dict(quantiles_plumes=quantiles_plumes, linewidth=1)

    check_plots(
        df=df.openscm.groupby_except("run")
        .quantile(get_quantiles(quantiles_plumes))
        .openscm.fix_index_name_after_groupby_quantile(),
        plot_kwargs=plot_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


@pytest.mark.parametrize(
    "quantile_over, hue_var, style_var, kwargs",
    (
        pytest.param(
            "scenario",
            "run",
            "variable",
            dict(warn_infer_y_label_with_multi_unit=False),
            id="single-var-with-style-var",
        ),
        pytest.param(
            ["scenario", "variable", "unit"],
            "run",
            None,
            dict(unit_var=None),
            id="multi-var-with-no-style-var",
        ),
    ),
)
def test_plot_plume_quantile_over(  # noqa: PLR0913
    quantile_over,
    hue_var,
    style_var,
    kwargs,
    tmp_path,
    image_regression,
    setup_pandas_accessor,
):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "W")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=56461),
    )

    method_kwargs = dict(
        quantile_over=quantile_over,
        quantiles_plumes=((0.5, 0.5), ((0.05, 0.95), 0.2)),
        hue_var=hue_var,
        style_var=style_var,
        **kwargs,
    )

    check_plots_incl_quantile_calculation(
        df=df,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


def test_plot_plume_extra_palette(
    tmp_path,
    image_regression,
    setup_pandas_accessor,
):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "W")),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=6543),
    )

    method_kwargs = dict(
        quantile_over="run",
        quantiles_plumes=((0.5, 0.5), ((0.05, 0.95), 0.2)),
        hue_var="scenario",
        palette={
            "scenario_0": "tab:green",
            "scenario_1": "tab:purple",
            "scenario_2": "tab:red",
            # Not df
            "scenario_3": "tab:orange",
        },
        style_var="variable",
        warn_infer_y_label_with_multi_unit=False,
    )

    check_plots_incl_quantile_calculation(
        df=df,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


def test_plot_plume_missing_from_palette(
    tmp_path,
    image_regression,
    setup_pandas_accessor,
):
    df = create_test_df(
        variables=(("variable_1", "K"),),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=85918),
    )

    palette = {
        "scenario_0": "tab:orange",
        # "scenario_1": "tab:blue",
        "scenario_2": "tab:blue",
    }
    method_kwargs = dict(
        quantile_over="run",
        quantiles_plumes=((0.5, 0.5), ((0.05, 0.95), 0.2)),
        hue_var="scenario",
        palette=palette,
    )

    check_plots_incl_quantile_calculation(
        df=df,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
        exp=pytest.warns(
            match=re.escape(
                "Some hue values are not in the user-supplied palette, "
                "they will be filled from the default colour cycler instead. "
                f"missing_from_user_supplied={['scenario_1']} "
                f"palette_user_supplied={palette}"
            )
        ),
    )


def test_plot_plume_extra_dashes(
    tmp_path,
    image_regression,
    setup_pandas_accessor,
):
    df = create_test_df(
        variables=(("variable_1", "W"), ("variable_2", "W")),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=6543),
    )

    method_kwargs = dict(
        quantile_over="run",
        quantiles_plumes=((0.5, 0.5), ((0.05, 0.95), 0.2)),
        hue_var="scenario",
        style_var="variable",
        dashes={
            # Not in df
            "variable_0": "-",
            "variable_1": "--",
            "variable_2": ":",
        },
    )

    check_plots_incl_quantile_calculation(
        df=df,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )


def test_plot_plume_missing_from_dashes(
    tmp_path,
    image_regression,
    setup_pandas_accessor,
):
    df = create_test_df(
        variables=(("variable_1", "W"), ("variable_2", "W")),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=85919),
    )

    dashes = {
        # "variable_1": "--",
        "variable_2": ":",
    }
    method_kwargs = dict(
        quantile_over="run",
        quantiles_plumes=((0.5, 0.5), ((0.05, 0.95), 0.2)),
        hue_var="scenario",
        style_var="variable",
        dashes=dashes,
    )

    check_plots_incl_quantile_calculation(
        df=df,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
        exp=pytest.warns(
            match=re.escape(
                "Some style values are not in the user-supplied dashes, "
                "they will be filled from the default dash cycler instead. "
                f"missing_from_user_supplied={['variable_1']} "
                f"dashes_user_supplied={dashes}"
            )
        ),
    )


@pytest.mark.parametrize(
    "quantiles, quantiles_plumes, exp",
    (
        pytest.param(
            [0.05, 0.5, 0.95],
            ((0.45, 0.5), ((0.05, 0.95), 0.2)),
            pytest.warns(match=r"Quantiles missing.*missing_quantiles=\[0.45\]"),
            id="missing-line-quantile",
        ),
        pytest.param(
            [0.05, 0.5, 0.95],
            ((0.5, 0.5), ((0.15, 0.95), 0.2)),
            pytest.warns(match=r"Quantiles missing.*missing_quantiles=\[0.15\]"),
            id="missing-plume-quantile",
        ),
        pytest.param(
            [0.05, 0.5, 0.95],
            ((0.5, 0.5), ((0.15, 0.85), 0.2)),
            pytest.warns(match=r"Quantiles missing.*missing_quantiles=\[0.15, 0.85\]"),
            id="missing-plumes-quantile",
        ),
    ),
)
def test_plot_plume_missing_quantiles(  # noqa: PLR0913
    quantiles, quantiles_plumes, exp, setup_pandas_accessor, image_regression, tmp_path
):
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=2,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=112345),
    )

    check_plots(
        df=df.openscm.groupby_except("run")
        .quantile(quantiles)
        .openscm.fix_index_name_after_groupby_quantile(),
        plot_kwargs=dict(quantiles_plumes=quantiles_plumes),
        image_regression=image_regression,
        tmp_path=tmp_path,
        exp=exp,
    )


def test_plot_plume_missing_multiple_quantiles(
    setup_pandas_accessor,
    image_regression,
    tmp_path,
    recwarn,
):
    quantiles = [0.25, 0.75]
    quantiles_plumes = ((0.5, 0.5), ((0.15, 0.85), 0.2), ((0.25, 0.75), 0.4))

    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=2,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
        rng=np.random.default_rng(seed=118844),
    )

    check_plots(
        df=df.openscm.groupby_except("run")
        .quantile(quantiles)
        .openscm.fix_index_name_after_groupby_quantile(),
        plot_kwargs=dict(quantiles_plumes=quantiles_plumes),
        image_regression=image_regression,
        tmp_path=tmp_path,
    )

    assert len(recwarn) == 16
    for w in recwarn:
        assert any(
            m in str(w.message)
            for m in (
                "missing_quantiles=[0.5]",
                "missing_quantiles=[0.15, 0.85]",
            )
        )


def test_plot_plume_option_passing(setup_pandas_accessor, image_regression, tmp_path):
    openscm_units = pytest.importorskip("openscm_units")
    openscm_units.unit_registry.setup_matplotlib(enable=True)

    df = create_test_df(
        variables=(("variable_1", "tCO2"), ("variable_2", "dtCO2")),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(2025.0, 2150.0),
        rng=np.random.default_rng(seed=85910),
    )

    pdf = (
        df.openscm.groupby_except("run")
        .quantile([0.1685321, 0.5, 0.8355321])
        .openscm.fix_index_name_after_groupby_quantile(new_name="percentile")
        .reset_index(["unit", "percentile"])
    )
    pdf["percentile"] *= 100.0
    pdf = pdf.rename({"unit": "units"}, axis="columns")
    pdf = pdf.set_index(["units", "percentile"], append=True)
    pdf.columns = pdf.columns.astype(float)

    def create_legend(ax, handles) -> None:
        ax.legend(handles=handles, loc="best", handlelength=4)

    plot_kwargs = dict(
        quantiles_plumes=((50.0, 1.0), ((16.85321, 83.55321), 0.3)),
        quantile_var="percentile",
        quantile_var_label="Percent",
        quantile_legend_round=3,
        hue_var="variable",
        hue_var_label="Var",
        palette={
            # Drop out to trigger warning below
            # "variable_1": "tab:green",
            "variable_2": "tab:purple",
        },
        warn_on_palette_value_missing=False,
        style_var="scenario",
        style_var_label="Scen",
        dashes={
            "scenario_0": "--",
            # Drop out to trigger warning below
            # "scenario_1": "-",
            "scenario_2": (0, (5, 3, 5, 1)),
        },
        warn_on_dashes_value_missing=False,
        linewidth=1.5,
        unit_var="units",
        unit_aware=openscm_units.unit_registry,
        time_units="month",
        x_label=None,  # let unit-awareness take over
        y_label=None,  # let unit-awareness take over
        # warn_infer_y_label_with_multi_unit tested elsewhere
        create_legend=create_legend,
        observed=False,
    )

    check_plots(
        df=pdf,
        plot_kwargs=plot_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )

    # Teardown
    openscm_units.unit_registry.setup_matplotlib(enable=False)


def test_plot_plume_after_calculating_quantiles_option_passing(
    setup_pandas_accessor, image_regression, tmp_path
):
    openscm_units = pytest.importorskip("openscm_units")
    openscm_units.unit_registry.setup_matplotlib(enable=True)

    df = create_test_df(
        variables=(("variable_1", "ctCO2"), ("variable_2", "dtCO2")),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(2025.0, 2150.0),
        rng=np.random.default_rng(seed=17583),
    )

    pdf = df.reset_index("unit")
    pdf = pdf.rename({"unit": "units"}, axis="columns")
    pdf = pdf.set_index("units", append=True)
    pdf.columns = pdf.columns.astype(float)

    def create_legend(ax, handles) -> None:
        ax.legend(handles=handles, loc="best", handlelength=4)

    method_kwargs = dict(
        quantile_over="run",
        quantiles_plumes=((0.5, 1.0), ((1.0 / 6.0, 5.0 / 6.0), 0.3)),
        quantile_var_label="Percent",
        quantile_legend_round=3,
        hue_var="variable",
        hue_var_label="Var",
        palette={
            # Drop out to trigger warning below
            # "variable_1": "tab:green",
            "variable_2": "tab:purple",
        },
        warn_on_palette_value_missing=False,
        style_var="scenario",
        style_var_label="Scen",
        dashes={
            "scenario_0": "--",
            # Drop out to trigger warning below
            # "scenario_1": "-",
            "scenario_2": (0, (5, 3, 5, 1)),
        },
        warn_on_dashes_value_missing=False,
        linewidth=1.5,
        unit_var="units",
        unit_aware=openscm_units.unit_registry,
        time_units="decade",
        x_label=None,  # let unit-awareness take over
        y_label=None,  # let unit-awareness take over
        # warn_infer_y_label_with_multi_unit tested elsewhere
        create_legend=create_legend,
        observed=False,
    )

    check_plots_incl_quantile_calculation(
        df=pdf,
        method_kwargs=method_kwargs,
        image_regression=image_regression,
        tmp_path=tmp_path,
    )

    # Teardown
    openscm_units.unit_registry.setup_matplotlib(enable=False)


@pytest.mark.parametrize(
    "unit_aware, variables",
    (
        pytest.param(
            True,
            (("variable_1", "cW"), ("variable_2", "mW")),
            id="default-unit-registry",
        ),
        pytest.param(
            getattr(openscm_units, "unit_registry", None),
            (("variable_1", "GtC"), ("variable_2", "GtCO2")),
            marks=pytest.mark.skipif(
                openscm_units is None, reason="openscm_units not installed"
            ),
            id="user-provided-unit-registry",
        ),
    ),
)
def test_plot_plume_unit_aware(
    unit_aware, variables, setup_pandas_accessor, image_regression, tmp_path
):
    """
    Make sure that we can do unit-aware plots

    In other words, even if the units are different,
    if they're compatible, they're plotted with the same units.
    """
    if isinstance(unit_aware, bool):
        pint = pytest.importorskip("pint")
        ur = pint.get_application_registry()
    else:
        ur = unit_aware

    ur.setup_matplotlib(enable=True)

    fig, ax = plt.subplots()

    res = (
        create_test_df(
            variables=variables,
            n_scenarios=3,
            n_runs=10,
            timepoints=np.arange(1950.0, 2050.0),
            rng=np.random.default_rng(seed=8888),
        )
        .openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile()
        .openscm.plot_plume(
            ax=ax,
            quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
            unit_aware=unit_aware,
            time_units="yr",
        )
    )

    assert res == ax

    out_file = tmp_path / "fig.png"
    plt.savefig(
        out_file,
        bbox_extra_artists=(ax.get_legend(),),
        bbox_inches="tight",
    )

    image_regression.check(out_file.read_bytes(), diff_threshold=0.01)

    plt.close()

    # Teardown
    ur.setup_matplotlib(enable=False)


def test_plot_plume_unit_aware_incompatible_units(setup_pandas_accessor):
    """
    Make sure that we can do unit-aware plots and errors are caught

    In other words, if the units are incompatible,
    then trying to plot on the same axes raises an error.
    """
    pint = pytest.importorskip("pint")

    ur = pint.get_application_registry()
    ur.setup_matplotlib(enable=True)

    df_m = create_test_df(
        variables=(("variable_1", "m"),),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 2050.0),
        rng=np.random.default_rng(seed=8888),
    )

    df_kg = create_test_df(
        variables=(("variable_2", "kg"),),
        n_scenarios=3,
        n_runs=10,
        timepoints=np.arange(1950.0, 2050.0),
        rng=np.random.default_rng(seed=8889),
    )

    _, ax = plt.subplots()

    # Plot first df
    df_m_grouped = (
        df_m.openscm.groupby_except("run")
        .quantile([0.05, 0.5, 0.95])
        .openscm.fix_index_name_after_groupby_quantile()
    )
    df_m_grouped.openscm.plot_plume(
        ax=ax,
        quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
        unit_aware=True,
        time_units="yr",
    )

    with pytest.raises(
        matplotlib.units.ConversionError,
        match=re.escape("Failed to convert value(s) to axis units"),
    ):
        # Plotting an incompatible unit raises.
        # Test plot_plume accesor (to double check argument passing)
        (
            df_kg.openscm.groupby_except("run")
            .quantile([0.05, 0.5, 0.95])
            .openscm.fix_index_name_after_groupby_quantile()
        ).openscm.plot_plume(
            ax=ax,
            quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
            unit_aware=True,
            time_units="yr",
        )

    with pytest.raises(
        matplotlib.units.ConversionError,
        match=re.escape("Failed to convert value(s) to axis units"),
    ):
        # Test that different x-axis also fails
        df_m_grouped.openscm.plot_plume(
            ax=ax,
            quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
            unit_aware=True,
            time_units="kg",
        )

    with pytest.raises(
        matplotlib.units.ConversionError,
        match=re.escape("Failed to convert value(s) to axis units"),
    ):
        # Test plot_plume_after_calculating_quantiles accesor
        # (to double check argument passing)
        df_kg.openscm.plot_plume_after_calculating_quantiles(
            ax=ax,
            quantile_over="run",
            quantiles_plumes=((0.5, 0.9), ((0.05, 0.95), 0.2)),
            unit_aware=True,
            time_units="yr",
        )

    # Teardown
    ur.setup_matplotlib(enable=False)


def test_get_default_colour_cycler_no_matplotlib():
    with patch.dict(sys.modules, {"matplotlib": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`get_default_colour_cycler` requires matplotlib to be installed"
            ),
        ):
            get_default_colour_cycler()


def test_extract_single_unit_raises_if_more_than_one():
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "W")),
        n_scenarios=1,
        n_runs=1,
        timepoints=np.arange(1950.0, 1965.0),
    )

    with pytest.raises(
        AssertionError,
        match=re.escape("['K', 'W']"),
    ):
        extract_single_unit(df, unit_var="unit")


def test_get_values_line_unit_aware_no_pint():
    df = create_test_df(
        variables=(("variable_1", "K"), ("variable_2", "K")),
        n_scenarios=5,
        n_runs=10,
        timepoints=np.arange(1950.0, 1965.0),
    )

    with patch.dict(sys.modules, {"pint": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`get_values_line(..., unit_aware=True, ...)` "
                "requires pint to be installed"
            ),
        ):
            get_values_line(df, unit_aware=True, unit_var="unit", time_units="yr")


def test_get_values_plume_unit_aware_no_pint(setup_pandas_accessor):
    df = (
        create_test_df(
            variables=(("variable_1", "K"), ("variable_2", "K")),
            n_scenarios=5,
            n_runs=10,
            timepoints=np.arange(1950.0, 1965.0),
        )
        .openscm.groupby_except("run")
        .quantile([0.25, 0.75])
        .openscm.fix_index_name_after_groupby_quantile()
    )

    with patch.dict(sys.modules, {"pint": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`get_values_plume(..., unit_aware=True, ...)` "
                "requires pint to be installed"
            ),
        ):
            get_values_plume(
                df,
                quantiles=[0.25, 0.75],
                quantile_var="quantile",
                unit_aware=True,
                unit_var="unit",
                time_units="yr",
            )


def test_generate_legend_handles_no_matplotlib():
    pp = PlumePlotter(
        lines=[],
        plumes=[],
        hue_var_label="a",
        style_var_label="b",
        quantile_var_label="c",
        palette={},
        dashes=None,
        x_label=None,
        y_label=None,
    )
    with patch.dict(sys.modules, {"matplotlib": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape(
                "`generate_legend_handles` requires matplotlib to be installed"
            ),
        ):
            pp.generate_legend_handles()


def test_plot_no_matplotlib():
    pp = PlumePlotter(
        lines=[],
        plumes=[],
        hue_var_label="a",
        style_var_label="b",
        quantile_var_label="c",
        palette={},
        dashes=None,
        x_label=None,
        y_label=None,
    )
    with patch.dict(sys.modules, {"matplotlib": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=re.escape("`plot(ax=None, ...)` requires matplotlib to be installed"),
        ):
            pp.plot()


def test_single_line_plotter_wrong_shape_y_vals():
    error_msg = re.escape(
        "`y_vals` must have the same shape as `x_vals`. "
        "Received `y_vals` with shape (2,) while `x_vals` has shape (3,)"
    )
    with pytest.raises(AssertionError, match=error_msg):
        SingleLinePlotter(
            np.array([1, 2, 3]),
            np.array([1, 2]),
            quantile=0.3,
            linewidth=1.0,
            linestyle="-",
            color="blue",
            alpha=0.3,
        )



"""
Test `pandas_openscm.index_manipulation.unify_index_levels`
"""

from __future__ import annotations

import numpy as np
import pandas as pd

from pandas_openscm.index_manipulation import unify_index_levels


def assert_index_equal_here(res: pd.MultiIndex, exp: pd.MultiIndex):
    """
    Assert that indexes are equal

    Special function here in case we need to do any pre-processing.
    """
    pd.testing.assert_index_equal(res, exp)


def test_unify_index_levels_already_matching():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2, 3),
            (4, 5, 6),
        ],
        names=["a", "b", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8, 9),
            (10, 11, 12),
        ],
        names=["a", "b", "c"],
    )

    # Should be no change
    exp_a = idx_a
    # Should be no change
    exp_b = idx_b

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_just_reordering():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2, 3),
            (4, 5, 6),
        ],
        names=["a", "b", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8, 9),
            (10, 11, 12),
        ],
        names=["c", "a", "b"],
    )

    # Should be no change
    exp_a = idx_a
    exp_b = idx_b.reorder_levels(idx_a.names)

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_a_within_b():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2),
            (4, 5),
        ],
        names=["a", "b"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8, 9),
            (10, 11, 12),
        ],
        names=["a", "b", "c"],
    )

    exp_a = pd.MultiIndex(
        levels=[[1, 4], [2, 5], np.array([], dtype=np.int64)],
        codes=[[0, 1], [0, 1], [-1, -1]],
        names=["a", "b", "c"],
    )

    # Should be no change
    exp_b = idx_b

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_a_within_b_skip_a_level():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2),
            (4, 5),
        ],
        names=["a", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8, 9),
            (10, 11, 12),
        ],
        names=["a", "b", "c"],
    )

    exp_a = pd.MultiIndex(
        levels=[[1, 4], [2, 5], np.array([], dtype=np.int64)],
        codes=[[0, 1], [0, 1], [-1, -1]],
        names=["a", "c", "b"],
    )

    exp_b = idx_b.reorder_levels(["a", "c", "b"])

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_a_outside_b():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2, 3),
            (4, 5, 6),
        ],
        names=["a", "b", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8),
            (10, 11),
        ],
        names=["a", "b"],
    )

    exp_b = pd.MultiIndex(
        levels=[[7, 10], [8, 11], np.array([], dtype=np.int64)],
        codes=[[0, 1], [0, 1], [-1, -1]],
        names=["a", "b", "c"],
    )

    # Should be no change
    exp_a = idx_a

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_a_outside_b_skip_level():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2, 3),
            (4, 5, 6),
        ],
        names=["a", "b", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8),
            (10, 11),
        ],
        names=["b", "c"],
    )

    exp_b = pd.MultiIndex(
        levels=[np.array([], dtype=np.int64), [7, 10], [8, 11]],
        codes=[[-1, -1], [0, 1], [0, 1]],
        names=["a", "b", "c"],
    )

    # Should be no change
    exp_a = idx_a

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_partial_intersecting():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2, 3),
            (4, 5, 6),
        ],
        names=["a", "b", "d"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (7, 8, 9),
            (10, 11, 12),
        ],
        names=["a", "b", "c"],
    )

    exp_a = pd.MultiIndex(
        levels=[[1, 4], [2, 5], [3, 6], np.array([], dtype=np.int64)],
        codes=[[0, 1], [0, 1], [0, 1], [-1, -1]],
        names=["a", "b", "d", "c"],
    )

    exp_b = pd.MultiIndex(
        levels=[[7, 10], [8, 11], np.array([], dtype=np.int64), [9, 12]],
        codes=[[0, 1], [0, 1], [-1, -1], [0, 1]],
        names=["a", "b", "d", "c"],
    )

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_non_intersecting():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (1, 2),
            (3, 4),
        ],
        names=["a", "c"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (5, 6),
            (7, 8),
        ],
        names=["b", "d"],
    )

    exp_a = pd.MultiIndex(
        levels=[
            [1, 3],
            [2, 4],
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
        ],
        codes=[[0, 1], [0, 1], [-1, -1], [-1, -1]],
        names=["a", "c", "b", "d"],
    )

    exp_b = pd.MultiIndex(
        levels=[
            np.array([], dtype=np.int64),
            np.array([], dtype=np.int64),
            [5, 7],
            [6, 8],
        ],
        codes=[[-1, -1], [-1, -1], [0, 1], [0, 1]],
        names=["a", "c", "b", "d"],
    )

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_ordering():
    idx_a = pd.MultiIndex.from_tuples(
        [
            (np.nan, 2, 1, np.nan),
        ],
        names=["a", "b", "c", "d"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            (11, 12),
            (5, 4),
            (7, 8),
        ],
        names=["b", "c"],
    )

    # Expect no change
    exp_a = idx_a

    exp_b = pd.MultiIndex(
        levels=[
            [],
            [11, 5, 7],
            [12, 4, 8],
            [],
        ],
        codes=[[-1, -1, -1], [0, 1, 2], [0, 1, 2], [-1, -1, -1]],
        names=["a", "b", "c", "d"],
    )

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)


def test_unify_index_levels_overlap():
    idx_a = pd.MultiIndex.from_tuples(
        [
            ("scenario_a", "model_1", "variable_1"),
            ("scenario_a", "model_2", "variable_1"),
            ("scenario_a", "model_1", "variable_2"),
            ("scenario_a", "model_2", "variable_2"),
            ("scenario_b", "model_1", "variable_1"),
            ("scenario_b", "model_2", "variable_1"),
            ("scenario_b", "model_1", "variable_2"),
            ("scenario_b", "model_2", "variable_2"),
        ],
        names=["scenario", "model", "variable"],
    )

    idx_b = pd.MultiIndex.from_tuples(
        [
            ("scenario_a", "variable_1"),
            ("scenario_a", "variable_1"),
            ("scenario_a", "variable_2"),
            ("scenario_a", "variable_2"),
            ("scenario_b", "variable_1"),
            ("scenario_b", "variable_1"),
            ("scenario_b", "variable_2"),
            ("scenario_b", "variable_2"),
        ],
        names=["scenario", "variable"],
    )

    # Expect no change
    exp_a = idx_a

    exp_b = pd.MultiIndex(
        levels=[
            ["scenario_a", "scenario_b"],
            [],
            ["variable_1", "variable_2"],
        ],
        codes=[
            [0, 0, 0, 0, 1, 1, 1, 1],
            [-1, -1, -1, -1, -1, -1, -1, -1],
            [0, 0, 1, 1, 0, 0, 1, 1],
        ],
        names=["scenario", "model", "variable"],
    )

    res_a, res_b = unify_index_levels(idx_a, idx_b)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)

    # Do it the other way around too
    res_b, res_a = unify_index_levels(idx_b, idx_a)

    exp_order = [*idx_b.names, *res_a.names.difference(idx_b.names)]
    exp_a = res_a.reorder_levels(exp_order)
    exp_b = res_b.reorder_levels(exp_order)

    assert_index_equal_here(res_a, exp_a)
    assert_index_equal_here(res_b, exp_b)



"""
Tests of speed of `pandas_openscm.OpenSCMDB`

These tests aim to catch places
where serialising data via OpenSCMDB
is much slower than directly using pandas.
These aren't perfect, but they're better than zero
for catching obvious mistakes.
"""

import time
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import FeatherDataBackend, FeatherIndexBackend, OpenSCMDB
from pandas_openscm.testing import create_test_df


@pytest.mark.parametrize(
    "groupby",
    (
        pytest.param(None, id="single-file"),
        pytest.param(["scenario"], id="scenario-groups"),
        pytest.param(["variable", "scenario"], id="scenario-variable-groups"),
    ),
)
def test_overhead(groupby, tmpdir):
    # Want lock time included in this test,
    # hence skip if not available
    pytest.importorskip("filelock")

    tmpdir = Path(tmpdir)

    feather_data_be = FeatherDataBackend()
    feather_index_be = FeatherIndexBackend()

    db = OpenSCMDB(
        db_dir=tmpdir / "db",
        backend_data=feather_data_be,
        backend_index=feather_index_be,
    )

    df = create_test_df(
        variables=[("Temperature", "K")],
        n_scenarios=150,
        n_runs=600,
        timepoints=np.arange(200.0),
    )

    pandas_out_dir = tmpdir / "pandas"
    pandas_out_dir.mkdir(parents=True)

    start_pandas_save = time.perf_counter()

    # Semi-replicate what the db has to do
    index = pd.DataFrame(
        np.full(df.shape[0], "path-to-somewhere"), index=df.index, columns=["file_path"]
    )
    index.to_feather(pandas_out_dir / "index.feather")

    if groupby is None:
        feather_data_be.save_data(df, pandas_out_dir / "data-pandas.feather")
    else:
        for i, (_, gdf) in enumerate(df.groupby(groupby)):
            gdf.to_feather(pandas_out_dir / f"{i}_data-pandas.feather")

    stop_pandas_save = time.perf_counter()
    time_pandas_save = stop_pandas_save - start_pandas_save

    start_db_save = time.perf_counter()

    db.db_dir.mkdir()
    db.save(df, groupby=groupby)

    stop_db_save = time.perf_counter()
    time_db_save = stop_db_save - start_db_save

    # Ok, particularly given how few files we're dealing with.
    # This is mainly about avoiding a factor of 10
    # (which was the difference we were getting in earlier implemenations).
    tol_save = 0.25
    tol_load = 0.2

    overhead = (time_db_save - time_pandas_save) / time_pandas_save
    assert overhead <= tol_save, f"Overhead is more than {tol_save*100}%"

    start_pandas_load = time.perf_counter()

    [pd.read_feather(f) for f in pandas_out_dir.glob("*.feather")]

    stop_pandas_load = time.perf_counter()
    time_pandas_load = stop_pandas_load - start_pandas_load

    start_db_load = time.perf_counter()

    db.load()

    stop_db_load = time.perf_counter()
    time_db_load = stop_db_load - start_db_load

    overhead = (time_db_load - time_pandas_load) / time_pandas_load
    assert overhead <= tol_load, f"Overhead is more than {tol_load*100}%"



"""
Loading of data from disk
"""

from __future__ import annotations

import concurrent.futures
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd

from pandas_openscm.db.interfaces import OpenSCMDBDataBackend, OpenSCMDBIndexBackend
from pandas_openscm.index_manipulation import (
    unify_index_levels_check_index_types,
    update_index_from_candidates,
)
from pandas_openscm.indexing import mi_loc
from pandas_openscm.parallelisation import (
    ParallelOpConfig,
    apply_op_parallel_progress,
)

if TYPE_CHECKING:
    import pandas.core.groupby.generic
    import pandas.core.indexes.frozen
    import pandas_indexing as pix


def load_data(  # noqa: PLR0913
    *,
    backend_data: OpenSCMDBDataBackend,
    db_index: pd.DataFrame,
    db_file_map: pd.Series[Path],  # type: ignore # pandas type hints confused about what they support
    selector: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector | None = None,
    out_columns_type: type | None = None,
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> pd.DataFrame:
    """
    Load data

    Parameters
    ----------
    backend_data
        Backend to use to load data from disk

    db_index
        Index of the database from which to load

    db_file_map
        File map of the database from which to load

    selector
        Selector to use to choose the data to load

    out_columns_type
        Type to set the output columns to.

        If not supplied, we don't set the output columns' type.

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the deletion?

        Only used if `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
        with the provided number of workers.
        A process pool seems to be the sensible default from our experimentation,
        but it is not a universally better choice.
        If you need something else because of how your database is set up,
        simply pass `parallel_op_config`
        rather than using the shortcut of passing `max_workers`.

        If not supplied, the loading is executed serially.

        Only used if `parallel_op_config` is `None`.

    Returns
    -------
    :
        Loaded data
    """
    if selector is None:
        index_to_load = db_index
    else:
        index_to_load = mi_loc(db_index, selector)

    files_to_load = (Path(v) for v in db_file_map[index_to_load["file_id"].unique()])
    loaded_l = load_data_files(
        files_to_load=files_to_load,
        backend_data=backend_data,
        parallel_op_config=parallel_op_config,
        progress=progress,
        max_workers=max_workers,
    )

    if backend_data.preserves_index and any(
        v.index.names != loaded_l[0].index.names for v in loaded_l
    ):
        base_idx = index_to_load.index[:1]
        for i in range(len(loaded_l)):
            new_index = unify_index_levels_check_index_types(
                base_idx, loaded_l[i].index
            )[1]
            loaded_l[i].index = new_index

    res = pd.concat(loaded_l)

    if not backend_data.preserves_index:
        index_names: pandas.core.indexes.frozen.FrozenList = index_to_load.index.names  # type: ignore # pandas type hints wrong
        res = update_index_from_candidates(res, index_names.difference({"file_id"}))

    # Look up only the indexes we want
    # just in case the data we loaded had more than we asked for
    # (because the files aren't saved with exactly the right granularity
    # for the query that has been requested).
    if selector is not None:
        res = mi_loc(res, selector)

    if out_columns_type is not None:
        res.columns = res.columns.astype(out_columns_type)

    return res


def load_data_files(
    *,
    files_to_load: Iterable[Path],
    backend_data: OpenSCMDBDataBackend,
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> tuple[pd.DataFrame, ...]:
    """
    Load a number of data files

    Parameters
    ----------
    files_to_load
        Files to load

    backend_data
        Data backend to use to load the files

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the deletion?

        Only used if `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
        with the provided number of workers.
        A process pool seems to be the sensible default from our experimentation,
        but it is not a universally better choice.
        If you need something else because of how your database is set up,
        simply pass `parallel_op_config`
        rather than using the shortcut of passing `max_workers`.

        If not supplied, the loading is executed serially.

        Only used if `parallel_op_config` is `None`.

    Returns
    -------
    :
        Loaded data from each file

    See Also
    --------
    [load_data][(m).], which loads data while applying selectors
    (in contrast to this function which just loads all data).
    """
    iterable_input: Iterable[Path] | list[Path] = files_to_load

    # Stick the whole thing in a try finally block so we shutdown
    # if we created the parallel pool, it is shut down even if interrupted.
    try:
        if parallel_op_config is None:
            parallel_op_config_use = ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc="File loading"),
                progress_parallel_submission_kwargs=dict(
                    desc="Submitting files to the parallel executor"
                ),
                max_workers=max_workers,
                # Process pool by default as basic tests suggest
                # that reading is CPU-bound.
                # See the docs for nuance though.
                parallel_pool_cls=concurrent.futures.ProcessPoolExecutor,
            )
        else:
            parallel_op_config_use = parallel_op_config

        if parallel_op_config_use.progress_results is not None:
            # Wrap in list to force the length to be available to any progress bar.
            # This might be the wrong decision in a weird edge case,
            # but it's convenient enough that I'm willing to take that risk
            iterable_input = list(iterable_input)

        res = apply_op_parallel_progress(
            func_to_call=backend_data.load_data,
            iterable_input=iterable_input,
            parallel_op_config=parallel_op_config_use,
        )

    finally:
        if parallel_op_config_use.executor_created_in_class_method:
            if parallel_op_config_use.executor is None:  # pragma: no cover
                # Should be impossible to get here
                raise AssertionError

            parallel_op_config_use.executor.shutdown()

    return res


def load_db_index(
    *,
    backend_index: OpenSCMDBIndexBackend,
    index_file: Path,
) -> pd.DataFrame:
    """
    Load database index from file

    Parameters
    ----------
    backend_index
        Backend to use to load the index

    index_file
        File from which to load the index

    Returns
    -------
    :
        Loaded index
    """
    db_index = backend_index.load_index(index_file)

    if not backend_index.preserves_index:
        db_index = db_index.set_index(
            db_index.columns.difference(["file_id"]).to_list()
        )

    return db_index


def load_db_file_map(
    *,
    backend_index: OpenSCMDBIndexBackend,
    file_map_file: Path,
) -> pd.Series[Path]:  # type: ignore # pandas type hints confused about what they support
    """
    Load database file map from file

    Parameters
    ----------
    backend_index
        Backend to use to load the index

    file_map_file
        File from which to load the file map

    Returns
    -------
    :
        Loaded file map
    """
    file_map_raw = backend_index.load_file_map(file_map_file)
    if not backend_index.preserves_index:
        file_map_indexed = file_map_raw.set_index("file_id")
    else:
        file_map_indexed = file_map_raw

    file_map = file_map_indexed["file_path"]

    return file_map


def convert_db_index_to_metadata(db_index: pd.DataFrame) -> pd.MultiIndex:
    """
    Convert a database index to metadata

    Parameters
    ----------
    db_index
        Database index

    Returns
    -------
    :
        Metadata
    """
    if not isinstance(db_index.index, pd.MultiIndex):  # pragma: no cover
        # Should be impossible to get here
        raise TypeError(db_index.index)

    res: pd.MultiIndex = db_index.index

    return res


def load_db_metadata(
    *,
    backend_index: OpenSCMDBIndexBackend,
    index_file: Path,
) -> pd.MultiIndex:
    """
    Load database metadata from file

    Parameters
    ----------
    backend_index
        Backend to use to load the index

    index_file
        File from which to load the index (from which the metadata is inferred)

    Returns
    -------
    :
        Loaded metadata
    """
    db_index = load_db_index(backend_index=backend_index, index_file=index_file)

    return convert_db_index_to_metadata(db_index)



"""
Re-useable fixtures etc. for tests

See https://docs.pytest.org/en/7.1.x/reference/fixtures.html#conftest-py-sharing-fixtures-across-multiple-files
"""

import pandas as pd
import pytest

import pandas_openscm.accessors


def pytest_runtest_setup(item):
    if any(
        item.iter_markers(name="superslow")
    ) and "superslow" not in item.config.getoption("-m"):
        pytest.skip("skip superslow by default")


@pytest.fixture(scope="session", autouse=True)
def pandas_terminal_width():
    # Set pandas terminal width so that doctests don't depend on terminal width.

    # We set the display width to 120 because examples should be short,
    # anything more than this is too wide to read in the source.
    pd.set_option("display.width", 120)

    # Display as many columns as you want (i.e. let the display width do the
    # truncation)
    pd.set_option("display.max_columns", 1000)


@pytest.fixture()
def setup_pandas_accessor() -> None:
    # Not parallel safe, but good enough
    pandas_openscm.register_pandas_accessor()

    yield None

    # Surprising and a bit annoying that there isn't a safer way to do this
    pd.DataFrame._accessors.discard("openscm")
    if hasattr(pd.DataFrame, "openscm"):
        del pd.DataFrame.openscm
    # In future
    # pd.Series._accessors.discard("openscm")
    # if hasattr(pd.Series, "openscm"):
    #     del pd.Series.openscm



"""
Helpers for working with pandas

Really these should either go into
[pandas_indexing](https://github.com/coroa/pandas-indexing)
or [pandas](https://github.com/pandas-dev/pandas)
long-term, but they're ok here for now.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    P = TypeVar("P", pd.DataFrame, pd.Series[Any])

    import pandas_indexing as pix


def multi_index_match(
    idx: pd.MultiIndex, locator: pd.MultiIndex
) -> np.typing.NDArray[np.bool]:
    """
    Perform a multi-index match

    This works, even if the levels of the locator are not the same
    as the levels of the index in which to match.

    Arguably, this should be moved to
    [pandas_indexing](https://github.com/coroa/pandas-indexing)
    or [pandas](https://github.com/pandas-dev/pandas).
    Relevant issues:

    - [pandas#55279](https://github.com/pandas-dev/pandas/issues/55279)
    - [pandas-indexing#64](https://github.com/coroa/pandas-indexing/issues/64)

    Parameters
    ----------
    idx
        Index in which to find matches

    locator
        Locator to use for finding matches

    Returns
    -------
    :
        Location of the rows in `idx` which are in `locator`.

    Raises
    ------
    KeyError
        `locator` has levels which are not in `idx`

    Examples
    --------
    >>> import pandas as pd
    >>> base = pd.MultiIndex.from_tuples(
    ...     (
    ...         ("ma", "sa", 1),
    ...         ("ma", "sb", 2),
    ...         ("mb", "sa", 1),
    ...         ("mb", "sb", 3),
    ...     ),
    ...     names=["model", "scenario", "id"],
    ... )
    >>>
    >>> # A locator that lines up with the multi-index levels exactly
    >>> loc_simple = pd.MultiIndex.from_tuples(
    ...     (
    ...         ("ma", "sa", 1),
    ...         ("mb", "sa", 1),
    ...     ),
    ...     names=["model", "scenario", "id"],
    ... )
    >>> multi_index_match(base, loc_simple)
    array([ True, False,  True, False])
    >>>
    >>> # A locator that lines up with the first level only
    >>> loc_first_level = pd.MultiIndex.from_tuples(
    ...     (("ma",),),
    ...     names=["model"],
    ... )
    >>> multi_index_match(base, loc_first_level)
    array([ True,  True, False, False])
    >>>
    >>> # A locator that lines up with the second level only
    >>> loc_first_level = pd.MultiIndex.from_tuples(
    ...     (("sa",),),
    ...     names=["scenario"],
    ... )
    >>> multi_index_match(base, loc_first_level)
    array([ True, False,  True, False])
    >>>
    >>> # A locator that lines up with the second and third level only
    >>> loc_first_level = pd.MultiIndex.from_tuples(
    ...     (("sb", 3),),
    ...     names=["scenario", "id"],
    ... )
    >>> multi_index_match(base, loc_first_level)
    array([False, False, False,  True])
    """
    try:
        idx_reordered: pd.MultiIndex = idx.reorder_levels(  # type: ignore # reorder_levels untyped
            [*locator.names, *idx.names.difference(locator.names)]  # type: ignore # pandas-stubs confused about difference
        )
    except KeyError as exc:
        unusable = locator.names.difference(idx.names)  # type: ignore # pandas-stubs confused about difference
        if unusable:
            msg = (
                f"The following levels in `locator` are not in `idx`: {unusable}. "
                f"{locator.names=} {idx.names=}"
            )
            raise KeyError(msg) from exc

        raise  # pragma: no cover

    return idx_reordered.isin(locator)


def multi_index_lookup(pandas_obj: P, locator: pd.MultiIndex) -> P:
    """
    Perform a multi-index look up

    For the problem this is solving, see [multi_index_match][(m)].

    Parameters
    ----------
    pandas_obj
        Pandas object in which to find matches

    locator
        Locator to use for finding matches

    Returns
    -------
    :
        Rows of `pandas_obj` that are in `locator`.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>>
    >>> base = pd.DataFrame(
    ...     data=np.arange(8).reshape((4, 2)),
    ...     columns=[2000, 2020],
    ...     index=pd.MultiIndex.from_tuples(
    ...         (
    ...             ("ma", "sa", 1),
    ...             ("ma", "sb", 2),
    ...             ("mb", "sa", 4),
    ...             ("mb", "sb", 3),
    ...         ),
    ...         names=["model", "scenario", "id"],
    ...     ),
    ... )
    >>>
    >>> # A locator that lines up with the second and third level only
    >>> loc_first_level = pd.MultiIndex.from_tuples(
    ...     (
    ...         ("sa", 1),
    ...         ("sb", 3),
    ...     ),
    ...     names=["scenario", "id"],
    ... )
    >>> multi_index_lookup(base, loc_first_level)
                       2000  2020
    model scenario id
    ma    sa       1      0     1
    mb    sb       3      6     7
    """
    if not isinstance(pandas_obj.index, pd.MultiIndex):
        msg = (
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            f"Received {type(pandas_obj.index)=}"
        )
        raise TypeError(msg)

    res = pandas_obj.loc[multi_index_match(pandas_obj.index, locator)]

    return res


def index_name_aware_match(
    idx: pd.MultiIndex, locator: pd.Index[Any]
) -> np.typing.NDArray[np.bool]:
    """
    Perform a match with an index, being aware of the index's name.

    This works, even if the index being looked up is not the first index.

    Parameters
    ----------
    idx
        Index in which to find matches

    locator
        Locator to use for finding matches

    Returns
    -------
    :
        Location of the rows in `idx` which are in `locator`, given `locator.name`.

    Examples
    --------
    >>> import pandas as pd
    >>>
    >>> base = pd.MultiIndex.from_tuples(
    ...     (
    ...         ("ma", "sa", 1),
    ...         ("ma", "sb", 2),
    ...         ("mb", "sa", 1),
    ...         ("mb", "sb", 3),
    ...     ),
    ...     names=["model", "scenario", "id"],
    ... )
    >>>
    >>> # A locator that lines up with the third level only
    >>> loc = pd.Index([1, 3], name="id")
    >>> index_name_aware_match(base, loc)
    array([ True, False,  True,  True])
    """
    res = idx.isin(locator.values, level=locator.name)

    return res


def index_name_aware_lookup(pandas_obj: P, locator: pd.Index[Any]) -> P:
    """
    Perform a look up with an index, being aware of the index's name.

    For the problem this is solving, see [index_name_aware_match][(m)].

    Parameters
    ----------
    pandas_obj
        Pandas object in which to find matches

    locator
        Locator to use for finding matches

    Returns
    -------
    :
        Rows of `pandas_obj` that are in `locator`, given `locator.name`.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>>
    >>> base = pd.DataFrame(
    ...     data=np.arange(8).reshape((4, 2)),
    ...     columns=[2000, 2020],
    ...     index=pd.MultiIndex.from_tuples(
    ...         (
    ...             ("ma", "sa", 1),
    ...             ("ma", "sb", 2),
    ...             ("mb", "sa", 4),
    ...             ("mb", "sb", 3),
    ...         ),
    ...         names=["model", "scenario", "id"],
    ...     ),
    ... )
    >>>
    >>> # A locator that lines up with the third level only
    >>> loc = pd.Index([1, 3], name="id")
    >>> index_name_aware_lookup(base, loc)
                       2000  2020
    model scenario id
    ma    sa       1      0     1
    mb    sb       3      6     7
    """
    if not isinstance(pandas_obj.index, pd.MultiIndex):
        msg = (
            "This function is only intended to be used "
            "when `df`'s index is an instance of `MultiIndex`. "
            f"Received {type(pandas_obj.index)=}"
        )
        raise TypeError(msg)

    return pandas_obj.loc[index_name_aware_match(pandas_obj.index, locator)]


def mi_loc(
    pandas_obj: P,
    locator: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector,
) -> P:
    """
    Select data, being slightly smarter than the default [pandas.DataFrame.loc][].

    Parameters
    ----------
    pandas_obj
        Pandas object on which to do the `.loc` operation

    locator
        Locator to apply

        If this is a multi-index, we use
        [multi_index_lookup][(m).] to ensure correct alignment.

        If this is an index that has a name,
        we use the name to ensure correct alignment.

    Returns
    -------
    :
        Selected data

    Notes
    -----
    If you have [pandas_indexing][] installed,
    you can get the same (perhaps even better) functionality
    using something like the following instead

    ```python
    ...
    pandas_obj.loc[pandas_indexing.isin(locator)]
    ...
    ```
    """
    if isinstance(locator, pd.MultiIndex):
        res: P = multi_index_lookup(pandas_obj, locator)

    elif isinstance(locator, pd.Index) and locator.name is not None:
        res = index_name_aware_lookup(pandas_obj, locator)

    else:
        res = pandas_obj.loc[locator]  # type: ignore

    return res



"""
CSV backend
"""

from __future__ import annotations

from pathlib import Path
from typing import Literal

import pandas as pd
from attrs import define


@define
class CSVDataBackend:
    """
    CSV data backend
    """

    ext: str = ".csv"
    """
    Extension to use with files saved by this backend.
    """

    @property
    def preserves_index(self) -> Literal[False]:
        """
        Whether this backend preserves the index of data upon (de-)serialisation
        """
        return False

    @staticmethod
    def load_data(data_file: Path) -> pd.DataFrame:
        """
        Load a data file

        Parameters
        ----------
        data_file
            File from which to load the data

        Returns
        -------
        :
            Loaded data
        """
        return pd.read_csv(data_file)

    @staticmethod
    def save_data(data: pd.DataFrame, data_file: Path) -> None:
        """
        Save data to disk

        Parameters
        ----------
        data
            Data to save

        data_file
            File in which to save the data
        """
        data.to_csv(data_file)


@define
class CSVIndexBackend:
    """
    CSV index backend
    """

    ext: str = ".csv"
    """
    Extension to use with files saved by this backend.
    """

    @property
    def preserves_index(self) -> Literal[False]:
        """
        Whether this backend preserves the `pd.MultiIndex` upon (de-)serialisation
        """
        return False

    @staticmethod
    def load_file_map(file_map_file: Path) -> pd.DataFrame:
        """
        Load the file map

        Parameters
        ----------
        file_map_file
            File from which to load the file map

        Returns
        -------
        :
            Loaded file map
        """
        return pd.read_csv(file_map_file)

    @staticmethod
    def load_index(index_file: Path) -> pd.DataFrame:
        """
        Load the index

        Parameters
        ----------
        index_file
            File from which to load the index

        Returns
        -------
        :
            Loaded index
        """
        return pd.read_csv(index_file)

    @staticmethod
    def save_file_map(
        file_map: pd.Series[Path],  # type: ignore # pandas confused about what it supports
        file_map_file: Path,
    ) -> None:
        """
        Save the file map to disk

        Parameters
        ----------
        file_map
            File map to save

        file_map_file
            File in which to save the file map
        """
        file_map.to_csv(file_map_file)

    @staticmethod
    def save_index(
        index: pd.DataFrame,
        index_file: Path,
    ) -> None:
        """
        Save the index to disk

        Parameters
        ----------
        index
            Index to save

        index_file
            File in which to save the index
        """
        index.to_csv(index_file)



"""
Tests of parallelisation and progress bars with `pandas_openscm.OpenSCMDB`
"""

from __future__ import annotations

import concurrent.futures
import multiprocessing
from contextlib import nullcontext
from functools import partial
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    EmptyDBError,
    FeatherDataBackend,
    FeatherIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.parallelisation import ParallelOpConfig
from pandas_openscm.testing import assert_frame_alike, create_test_df

tqdm_auto = pytest.importorskip("tqdm.auto")


@pytest.mark.slow
@pytest.mark.parametrize(
    "max_workers",
    (
        pytest.param(None, id="serial"),
        pytest.param(1, id="one-worker"),
        pytest.param(4, id="four-workers"),
    ),
)
@pytest.mark.parametrize(
    "progress",
    (pytest.param(False, id="no-progress"), pytest.param(True, id="progress")),
)
def test_save_load_overwrite_delete_parallel(tmpdir, progress, max_workers):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        # Have to use a file backed thing here
        # as memory backed stuff doesn't work in parallel.
        backend_data=FeatherDataBackend(),
        backend_index=FeatherIndexBackend(),
    )

    df = create_test_df(
        n_scenarios=15,
        variables=[
            ("variable_1", "kg"),
            ("variable_2", "Mt"),
            ("variable_3", "m"),
        ],
        n_runs=600,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    db.save(
        df, groupby=["scenario", "variable"], progress=progress, max_workers=max_workers
    )

    loaded = db.load(
        out_columns_type=df.columns.dtype, progress=progress, max_workers=max_workers
    )
    assert_frame_alike(loaded, df)

    partial_rewrite_loc = pix.isin(
        scenario=["scenario_1"], variable=["variable_1"], run=range(20)
    )
    partial_rewrite = df.loc[partial_rewrite_loc]
    db.save(
        partial_rewrite,
        groupby=["scenario", "variable"],
        progress=progress,
        max_workers=max_workers,
        allow_overwrite=True,
        warn_on_partial_overwrite=False,
    )

    exp_after_rewrite = pd.concat(
        [
            df.loc[~partial_rewrite_loc],
            partial_rewrite,
        ]
    )
    loaded = db.load(
        out_columns_type=df.columns.dtype, progress=progress, max_workers=max_workers
    )
    assert_frame_alike(loaded, exp_after_rewrite)

    db.delete(progress=progress, max_workers=max_workers)

    with pytest.raises(EmptyDBError):
        db.load_metadata()

    with pytest.raises(EmptyDBError):
        db.load()


@pytest.mark.slow
@pytest.mark.parametrize(
    "executor_ctx_manager, executor_ctx_manager_kwargs",
    (
        pytest.param(nullcontext, dict(enter_result=None), id="serial"),
        pytest.param(
            concurrent.futures.ThreadPoolExecutor,
            dict(max_workers=4),
            id="inject-threading",
        ),
        # If you use fork context here, pytest may hang.
        # This is a gotcha of parallelism.
        pytest.param(
            concurrent.futures.ProcessPoolExecutor,
            dict(max_workers=4, mp_context=multiprocessing.get_context("spawn")),
            id="inject-process",
        ),
    ),
)
@pytest.mark.parametrize(
    "save_progress_kwargs, load_progress",
    (
        pytest.param(
            dict(
                progress_grouping=partial(tqdm_auto.tqdm, desc="Custom grouping"),
                parallel_op_config_save=ParallelOpConfig(
                    progress_results=partial(tqdm_auto.tqdm, desc="Custom save"),
                    progress_parallel_submission=partial(
                        tqdm_auto.tqdm, desc="Custom save submit"
                    ),
                ),
                parallel_op_config_delete=ParallelOpConfig(
                    progress_results=partial(tqdm_auto.tqdm, desc="Custom save"),
                    progress_parallel_submission=partial(
                        tqdm_auto.tqdm, desc="Custom save submit"
                    ),
                ),
                parallel_op_config_rewrite=ParallelOpConfig(
                    progress_results=partial(tqdm_auto.tqdm, desc="Custom save"),
                    progress_parallel_submission=partial(
                        tqdm_auto.tqdm, desc="Custom save submit"
                    ),
                ),
            ),
            ParallelOpConfig(
                progress_results=partial(tqdm_auto.tqdm, desc="Custom load"),
                progress_parallel_submission=partial(
                    tqdm_auto.tqdm, desc="Custom load submit"
                ),
            ),
            id="custom-progress",
        ),
    ),
)
def test_save_load_delete_parallel_custom_progress(
    tmpdir,
    save_progress_kwargs,
    load_progress,
    executor_ctx_manager,
    executor_ctx_manager_kwargs,
):
    pix = pytest.importorskip("pandas_indexing")

    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        # Have to use a file backed thing here
        # as memory backed stuff doesn't work in parallel.
        backend_data=FeatherDataBackend(),
        backend_index=FeatherIndexBackend(),
    )

    df = create_test_df(
        n_scenarios=15,
        variables=[
            ("variable_1", "kg"),
            ("variable_2", "Mt"),
            ("variable_3", "m"),
        ],
        n_runs=600,
        timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
    )

    with executor_ctx_manager(**executor_ctx_manager_kwargs) as executor:
        if isinstance(executor, concurrent.futures.Executor):
            for k, v in save_progress_kwargs.items():
                if k.startswith("parallel_op_config"):
                    v.executor = executor

        elif executor is None:
            pass

        else:
            raise NotImplementedError(executor)

        db.save(df, groupby=["scenario", "variable"], **save_progress_kwargs)

        loaded = db.load(
            out_columns_type=df.columns.dtype, parallel_op_config=load_progress
        )
        assert_frame_alike(loaded, df)

        partial_rewrite_loc = pix.isin(
            scenario=["scenario_1"], variable=["variable_1"], run=range(20)
        )
        partial_rewrite = df.loc[partial_rewrite_loc]
        db.save(
            partial_rewrite,
            groupby=["scenario", "variable"],
            **save_progress_kwargs,
            allow_overwrite=True,
            warn_on_partial_overwrite=False,
        )

        exp_after_rewrite = pd.concat(
            [
                df.loc[~partial_rewrite_loc],
                partial_rewrite,
            ]
        )
        loaded = db.load(
            out_columns_type=df.columns.dtype, parallel_op_config=load_progress
        )
        assert_frame_alike(loaded, exp_after_rewrite)

        db.delete(parallel_op_config=save_progress_kwargs["parallel_op_config_delete"])

    with pytest.raises(EmptyDBError):
        db.load_metadata()

    with pytest.raises(EmptyDBError):
        db.load()



"""
Database reader

A small optimisation to allow for a reader that holds the index in memory,
rather than loading it from disk on every operation.
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

import pandas as pd
from attrs import define, field

from pandas_openscm.db.interfaces import OpenSCMDBDataBackend
from pandas_openscm.db.loading import convert_db_index_to_metadata, load_data
from pandas_openscm.parallelisation import (
    ParallelOpConfig,
)

if TYPE_CHECKING:
    from types import TracebackType

    import filelock
    import pandas_indexing as pix


@define
class OpenSCMDBReader:
    """
    Reader for reading data out of a database created with `OpenSCMDB`

    Holds the database file map and index in memory,
    so this can be faster for repeated reads.
    """

    backend_data: OpenSCMDBDataBackend = field(kw_only=True)
    """
    The backend for reading data from disk
    """

    db_file_map: pd.Series[Path] = field(kw_only=True)  # type: ignore # pandas type hints confused about what they support
    """
    The file map of the database from which we are reading.
    """

    db_index: pd.DataFrame = field(kw_only=True)
    """
    The index of the database from which we are reading.
    """

    lock: filelock.BaseFileLock | None = field(kw_only=True)
    """
    Lock for the database from which data is being read

    If `None`, we don't hold the lock and automatic locking is not enabled.
    """

    def __enter__(self) -> OpenSCMDBReader:
        """
        If the reader has a lock, acquire it
        """
        if self.lock is not None:
            self.lock.acquire()

        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        If the reader has a lock, release it
        """
        if self.lock is not None:
            self.lock.release()

    @property
    def metadata(self) -> pd.MultiIndex:
        """
        Database's metadata
        """
        return convert_db_index_to_metadata(db_index=self.db_index)

    def load(
        self,
        selector: pd.Index[Any] | pd.MultiIndex | pix.selectors.Selector | None = None,
        *,
        out_columns_type: type | None = None,
        parallel_op_config: ParallelOpConfig | None = None,
        progress: bool = False,
        max_workers: int | None = None,
    ) -> pd.DataFrame:
        """
        Load data

        Parameters
        ----------
        selector
            Selector to use to choose the data to load

        out_columns_type
            Type to set the output columns to.

            If not supplied, we don't set the output columns' type.

        parallel_op_config
            Configuration for executing the operation in parallel with progress bars

            If not supplied, we use the values of `progress` and `max_workers`.

        progress
            Should progress bar(s) be used to display the progress of the deletion?

            Only used if `parallel_op_config` is `None`.

        max_workers
            Maximum number of workers to use for parallel processing.

            If supplied, we create an instance of
            [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
            with the provided number of workers.
            A process pool seems to be the sensible default from our experimentation,
            but it is not a universally better choice.
            If you need something else because of how your database is set up,
            simply pass `parallel_op_config`
            rather than using the shortcut of passing `max_workers`.

            If not supplied, the loading is executed serially.

            Only used if `parallel_op_config` is `None`.

        Returns
        -------
        :
            Loaded data
        """
        return load_data(
            backend_data=self.backend_data,
            db_index=self.db_index,
            db_file_map=self.db_file_map,
            selector=selector,
            out_columns_type=out_columns_type,
            parallel_op_config=parallel_op_config,
            progress=progress,
            max_workers=max_workers,
        )



"""
Test `pandas_openscm.index_manipulation.convert_index_to_category_index`
"""

from __future__ import annotations

import numpy as np

from pandas_openscm.index_manipulation import convert_index_to_category_index
from pandas_openscm.testing import create_test_df


def run_checks(res, start):
    # Check that columns are now all category types
    for idx_lvl in res.index.names:
        # Check that we didn't start with categories or mangle the original DataFrame
        assert (
            str(start.index.get_level_values(idx_lvl).dtype) != "category"
        ), "Testing nothing"

        assert str(res.index.get_level_values(idx_lvl).dtype) == "category"

    # Check that memory usage went down.
    # Unclear to me why this doesn't work if I try and use memory_usage
    # on the index directly, without casting to frame first.
    assert (
        res.index.to_frame(index=False).memory_usage().sum()
        / start.index.to_frame(index=False).memory_usage().sum()
    ) < 0.5


def test_to_category_index():
    units = ["Mt", "kg", "W"]

    # Biggish DataFrame
    start = create_test_df(
        variables=[(f"variable_{i}", units[i % len(units)]) for i in range(25)],
        n_scenarios=30,
        n_runs=60,
        timepoints=np.arange(1750.0, 2100.0 + 1.0),
    )

    res = convert_index_to_category_index(start)

    run_checks(res, start)


def test_to_category_index_series():
    units = ["Mt", "kg", "W"]

    # Biggish DataFrame
    start = create_test_df(
        variables=[(f"variable_{i}", units[i % len(units)]) for i in range(25)],
        n_scenarios=30,
        n_runs=60,
        timepoints=np.arange(1750.0, 1752.0),
    )[1751.0]

    res = convert_index_to_category_index(start)

    run_checks(res, start)


def test_accessor(setup_pandas_accessor):
    units = ["Mt", "kg", "W"]

    # Biggish DataFrame
    start = create_test_df(
        variables=[(f"variable_{i}", units[i % len(units)]) for i in range(25)],
        n_scenarios=30,
        n_runs=60,
        timepoints=np.arange(1750.0, 2100.0 + 1.0),
    )

    res = start.openscm.to_category_index()

    run_checks(res, start)



"""
Basic unit tests of `pandas_openscm.database.netcdf`
"""

from __future__ import annotations

import sys
from unittest.mock import patch

import pandas as pd
import pytest

from pandas_openscm.db import netCDFDataBackend, netCDFIndexBackend
from pandas_openscm.db.netcdf import metadata_df_to_xr
from pandas_openscm.exceptions import MissingOptionalDependencyError


@pytest.mark.parametrize(
    "to_call, exp_name, args",
    (
        (netCDFDataBackend().load_data, "netCDFBackend.load_data", ["file_path"]),
        (
            netCDFDataBackend().save_data,
            "netCDFBackend.save_data",
            ["data", "file_path"],
        ),
        (
            netCDFIndexBackend().load_file_map,
            "netCDFBackend.load_file_map",
            ["file_path"],
        ),
        (
            netCDFIndexBackend().load_index,
            "netCDFBackend.load_index",
            ["file_path"],
        ),
        (
            netCDFIndexBackend().save_file_map,
            "netCDFBackend.save_file_map",
            ["file_map", "file_path"],
        ),
        (
            netCDFIndexBackend().save_index,
            "metadata_df_to_xr",
            [pd.DataFrame(), "file_path"],
        ),
        (
            metadata_df_to_xr,
            "metadata_df_to_xr",
            ["metadata"],
        ),
    ),
)
def test_xarray_not_available(to_call, exp_name, args):
    with patch.dict(sys.modules, {"xarray": None}):
        with pytest.raises(
            MissingOptionalDependencyError,
            match=(f"`{exp_name}` requires xarray to be installed"),
        ):
            to_call(*args)



"""
Functionality for saving data
"""

from __future__ import annotations

import concurrent.futures
from collections.abc import Iterable
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

import numpy as np
import pandas as pd
from attrs import define

from pandas_openscm.db.interfaces import OpenSCMDBDataBackend, OpenSCMDBIndexBackend
from pandas_openscm.index_manipulation import (
    unify_index_levels_check_index_types,
)
from pandas_openscm.parallelisation import (
    ParallelOpConfig,
    ProgressLike,
    apply_op_parallel_progress,
    get_tqdm_auto,
)

if TYPE_CHECKING:
    import pandas.core.groupby.generic
    import pandas.core.indexes.frozen


class DBFileType(Enum):
    """
    Type of a database file

    Really just a helper for [save_data][(m).]
    """

    DATA = auto()
    INDEX = auto()
    FILE_MAP = auto()


@define
class SaveAction:
    """A database save action"""

    info: pd.DataFrame | pd.Series[Any]
    """Information to save"""

    info_kind: DBFileType
    """The kind of information that this is"""

    backend: OpenSCMDBDataBackend | OpenSCMDBIndexBackend
    """Backend to use to save the data to disk"""

    save_path: Path
    """Path in which to save the information"""


def save_data(  # noqa: PLR0913
    data: pd.DataFrame,
    *,
    backend_data: OpenSCMDBDataBackend,
    get_new_data_file_path: Callable[[int], Path],
    backend_index: OpenSCMDBIndexBackend,
    index_file: Path,
    file_map_file: Path,
    index_non_data: pd.DataFrame | None = None,
    file_map_non_data: pd.Series[Path] | None = None,  # type: ignore # pandas type hints doesn't know what it supports
    min_file_id: int = 0,
    groupby: list[str] | None = None,
    progress_grouping: ProgressLike | None = None,
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> None:
    """
    Save data

    Parameters
    ----------
    data
        Data to save

    db
        Database in which to save the data

    index_non_data
        Index that is already in the database but isn't related to data.

        If supplied, this is combined with the index generated for `data`
        before we write the database's index.

    file_map_non_data
        File map that is already in the database but isn't related to data.

        If supplied, this is combined with the file map generated for `data`
        before we write the database's file map.

    min_file_id
        Minimum file ID to assign to save data chunks

    groupby
        Metadata columns to use to group the data.

        If not supplied, we save all the data in a single file.

    progress_grouping
        Progress bar to use when grouping the data

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the saving?

        Only used if `progress_grouping` is `None` or `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
        with the provided number of workers.
        A process pool seems to be the sensible default from our experimentation,
        but it is not a universally better choice.
        If you need something else because of how your database is set up,
        simply pass `parallel_op_config`
        rather than using the shortcut of passing `max_workers`.

        If not supplied, the saving is executed serially.

        Only used if `parallel_op_config` is `None`.
    """
    if groupby is None:
        # Write as a single file
        grouper: (
            Iterable[tuple[tuple[Any, ...], pd.DataFrame]]
            | pandas.core.groupby.generic.DataFrameGroupBy[
                # Switch to the below
                # when we switch mypy checks to using python > 3.9
                # tuple[Any, ...], Literal[True]
                tuple[Any, ...]
            ]
        ) = [((None,), data)]
    else:
        # Only want combos that are actually in the data
        grouper = data.groupby(groupby, observed=True)

    if progress_grouping or progress:
        if progress_grouping is None:
            progress_grouping = get_tqdm_auto(desc="Grouping data to save")

        grouper = progress_grouping(grouper)

    if index_non_data is None:
        index_non_data_unified_index = None
    else:
        unified_index = unify_index_levels_check_index_types(
            index_non_data.index, data.index[:1]
        )[0]
        index_non_data_unified_index = pd.DataFrame(
            index_non_data.values,
            index=unified_index,
            columns=index_non_data.columns,
        )

    write_groups_l = []
    index_data_out_l = []
    file_map_out = pd.Series(
        [],
        index=pd.Index([], name="file_id"),
        name="file_path",
    )
    for increment, (_, df) in enumerate(grouper):
        file_id = min_file_id + increment

        new_file_path = get_new_data_file_path(file_id)

        file_map_out.loc[file_id] = new_file_path  # type: ignore # pandas types confused about what they support
        if index_non_data_unified_index is None:
            df_index_unified = df.index
        else:
            _, df_index_unified = unify_index_levels_check_index_types(
                index_non_data_unified_index.index[:1], df.index
            )

        index_data_out_l.append(
            pd.DataFrame(
                np.full(df.index.shape[0], file_id),
                index=df_index_unified,
                columns=["file_id"],
            )
        )

        write_groups_l.append(
            SaveAction(
                info=df,
                info_kind=DBFileType.DATA,
                backend=backend_data,
                save_path=new_file_path,
            )
        )

    if index_non_data_unified_index is None:
        index_out = pd.concat(index_data_out_l)
    else:
        index_out = pd.concat([index_non_data_unified_index, *index_data_out_l])

    if file_map_non_data is not None:
        file_map_out = pd.concat([file_map_non_data, file_map_out])  # type: ignore # pandas-stubs confused

    # Write the index first as it can be slow if very big
    write_groups_l.insert(
        0,
        SaveAction(
            info=index_out,
            info_kind=DBFileType.INDEX,
            backend=backend_index,
            save_path=index_file,
        ),
    )
    # Write the file map last, it is almost always cheapest
    write_groups_l.append(
        SaveAction(
            info=file_map_out,
            info_kind=DBFileType.FILE_MAP,
            backend=backend_index,
            save_path=file_map_file,
        )
    )

    save_files(
        write_groups_l,
        parallel_op_config=parallel_op_config,
        progress=progress,
        max_workers=max_workers,
    )


def save_files(
    save_actions: Iterable[SaveAction],
    parallel_op_config: ParallelOpConfig | None = None,
    progress: bool = False,
    max_workers: int | None = None,
) -> None:
    """
    Save database information to disk

    Parameters
    ----------
    save_actions
        Iterable of save actions

    parallel_op_config
        Configuration for executing the operation in parallel with progress bars

        If not supplied, we use the values of `progress` and `max_workers`.

    progress
        Should progress bar(s) be used to display the progress of the deletion?

        Only used if `parallel_op_config` is `None`.

    max_workers
        Maximum number of workers to use for parallel processing.

        If supplied, we create an instance of
        [concurrent.futures.ProcessPoolExecutor](https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ProcessPoolExecutor)
        with the provided number of workers.
        A process pool seems to be the sensible default from our experimentation,
        but it is not a universally better choice.
        If you need something else because of how your database is set up,
        simply pass `parallel_op_config`
        rather than using the shortcut of passing `max_workers`.

        If not supplied, the saving is executed serially.

        Only used if `parallel_op_config` is `None`.
    """
    iterable_input: Iterable[SaveAction] | list[SaveAction] = save_actions

    # Stick the whole thing in a try finally block so we shutdown
    # the parallel pool, even if interrupted, if we created it.
    try:
        if parallel_op_config is None:
            parallel_op_config_use = ParallelOpConfig.from_user_facing(
                progress=progress,
                progress_results_kwargs=dict(desc="File saving"),
                progress_parallel_submission_kwargs=dict(
                    desc="Submitting files to the parallel executor"
                ),
                max_workers=max_workers,
                # Process pool by default as basic tests suggest
                # that writing is CPU-bound.
                # See the docs for nuance though.
                parallel_pool_cls=concurrent.futures.ProcessPoolExecutor,
            )
        else:
            parallel_op_config_use = parallel_op_config

        if parallel_op_config_use.progress_results is not None:
            # Wrap in list to force the length to be available to any progress bar.
            # This might be the wrong decision in a weird edge case,
            # but it's convenient enough that I'm willing to take that risk
            iterable_input = list(iterable_input)

        apply_op_parallel_progress(
            func_to_call=save_file,
            iterable_input=iterable_input,
            parallel_op_config=parallel_op_config_use,
        )

    finally:
        if parallel_op_config_use.executor_created_in_class_method:
            if parallel_op_config_use.executor is None:  # pragma: no cover
                # Should be impossible to get here
                raise AssertionError

            parallel_op_config_use.executor.shutdown()


def save_file(save_action: SaveAction) -> None:
    """
    Save a file to disk

    Parameters
    ----------
    save_action
        Save action to perform
    """
    if save_action.info_kind == DBFileType.DATA:
        if isinstance(save_action.info, pd.Series) or isinstance(
            save_action.backend, OpenSCMDBIndexBackend
        ):  # pragma: no cover
            # Should be impossible to get here
            raise TypeError

        save_action.backend.save_data(save_action.info, save_action.save_path)

    elif save_action.info_kind == DBFileType.INDEX:
        if isinstance(save_action.info, pd.Series) or isinstance(
            save_action.backend, OpenSCMDBDataBackend
        ):  # pragma: no cover
            # Should be impossible to get here
            raise TypeError

        save_action.backend.save_index(
            index=save_action.info,
            index_file=save_action.save_path,
        )

    elif save_action.info_kind == DBFileType.FILE_MAP:
        if isinstance(save_action.info, pd.DataFrame) or isinstance(
            save_action.backend, OpenSCMDBDataBackend
        ):  # pragma: no cover
            # Should be impossible to get here
            raise TypeError

        save_action.backend.save_file_map(
            file_map=save_action.info,
            file_map_file=save_action.save_path,
        )

    else:
        raise NotImplementedError(save_action.info_kind)



"""
Tests of deleting with `pandas_openscm.OpenSCMDB`
"""

from __future__ import annotations

import contextlib
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pandas_openscm.db import (
    EmptyDBError,
    InMemoryDataBackend,
    InMemoryIndexBackend,
    OpenSCMDB,
)
from pandas_openscm.testing import (
    create_test_df,
)


def test_delete(tmpdir):
    db = OpenSCMDB(
        db_dir=Path(tmpdir),
        backend_data=InMemoryDataBackend(),
        backend_index=InMemoryIndexBackend(),
        index_file_lock=contextlib.nullcontext(),  # not used
    )

    db.save(
        create_test_df(
            n_scenarios=10,
            variables=[("variable_1", "kg"), ("variable_2", "Mt"), ("variable_3", "m")],
            n_runs=3,
            timepoints=np.array([2010.0, 2020.0, 2025.0, 2030.0]),
        )
    )

    assert isinstance(db.load(), pd.DataFrame)

    db.delete()

    with pytest.raises(EmptyDBError):
        db.load_metadata()

    with pytest.raises(EmptyDBError):
        db.load()



"""
Testing helpers

Placed here to avoid putting `__init__.py` files in our `tests` directory,
see details here: https://docs.pytest.org/en/stable/explanation/goodpractices.html#which-import-mode.
Also see here: https://docs.pytest.org/en/stable/explanation/pythonpath.html#pytest-import-mechanisms-and-sys-path-pythonpath.
"""

from __future__ import annotations

import itertools
from collections.abc import Collection
from typing import TYPE_CHECKING, Any, cast

import numpy as np
import pandas as pd

from pandas_openscm.db import DATA_BACKENDS, INDEX_BACKENDS
from pandas_openscm.db.rewriting import MovePlan
from pandas_openscm.exceptions import MissingOptionalDependencyError

if TYPE_CHECKING:
    import pytest


def get_db_data_backends() -> tuple[type[object], ...]:
    return tuple(v[1] for v in DATA_BACKENDS.options)


def get_db_index_backends() -> tuple[type[object], ...]:
    return tuple(v[1] for v in INDEX_BACKENDS.options)


def get_parametrized_db_data_backends() -> pytest.MarkDecorator:
    try:
        import pytest
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_parametrized_db_data_backends", requirement="pytest"
        ) from exc

    return pytest.mark.parametrize(
        "db_data_backend",
        tuple(
            pytest.param(db_data_format, id=f"backend_data_{short_name}")
            for short_name, db_data_format in DATA_BACKENDS.options
        ),
    )


def get_parametrized_db_index_backends() -> pytest.MarkDecorator:
    try:
        import pytest
    except ImportError as exc:
        raise MissingOptionalDependencyError(
            "get_parametrized_db_index_backends", requirement="pytest"
        ) from exc

    return pytest.mark.parametrize(
        "db_index_backend",
        tuple(
            pytest.param(db_data_format, id=f"backend_index_{short_name}")
            for short_name, db_data_format in INDEX_BACKENDS.options
        ),
    )


def assert_frame_alike(
    res: pd.DataFrame, exp: pd.DataFrame, check_like: bool = True, **kwargs: Any
) -> None:
    """
    Assert that two [pd.DataFrame][pandas.DataFrame] are alike

    Here, alike means that they have the same data,
    just potentially not in the same order.
    This includes the order of index levels, which may also differ.

    Parameters
    ----------
    res
        Result to check

    exp
        Expected result

    check_like
        Passed to [assert_frame_equal][pandas.testing.assert_frame_equal]

    **kwargs
        Passed to [assert_frame_equal][pandas.testing.assert_frame_equal]
    """
    pd.testing.assert_frame_equal(
        res.reorder_levels(exp.index.names),
        exp,
        check_like=check_like,
        **kwargs,
    )


def create_test_df(
    *,
    variables: Collection[tuple[str, str]],
    n_scenarios: int,
    n_runs: int,
    timepoints: np.typing.NDArray[np.floating[Any]],
    rng: np.random.Generator | None = None,
) -> pd.DataFrame:
    """
    Create a [pd.DataFrame][pandas.DataFrame] to use in testing

    This uses the idea of simple climate model runs,
    where you have a number of scenarios,
    each of which has a number of variables
    from a number of different model runs
    with output for a number of different time points.

    The result will contain all combinations of scenarios,
    variables and runs,
    with the units being defined by each variable.

    Parameters
    ----------
    variables
        Variables and their units to create

    n_scenarios
        Number of scenarios to create.

        These are simply incremented with their number.

    n_runs
        Number of runs to create.

        These are simply numbered.

    timepoints
        Time points to use with the data.

    rng
        Random number generator to use.

        If not supplied, we create an instance using
        [numpy.random.default_rng](https://numpy.org/doc/stable/reference/random/generator.html#numpy.random.default_rng).

    Returns
    -------
    :
        Generated test [pd.DataFrame][pandas.DataFrame].
    """
    if rng is None:
        rng = np.random.default_rng()

    idx = pd.MultiIndex.from_frame(
        pd.DataFrame(
            (
                (s, v_info[0], r, v_info[1])
                for s, v_info, r in itertools.product(
                    [f"scenario_{i}" for i in range(n_scenarios)],
                    variables,
                    [i for i in range(n_runs)],
                )
            ),
            columns=["scenario", "variable", "run", "unit"],
        )
    )

    n_variables = len(variables)
    n_ts = n_scenarios * n_variables * n_runs

    # Give the data a bit of structure so it looks different when plotted.
    values = 50.0 * np.linspace(0.3, 1, n_ts)[:, np.newaxis] * np.linspace(
        0, 1, timepoints.size
    )[np.newaxis, :] + rng.random((n_ts, timepoints.size))

    df = pd.DataFrame(
        values,
        columns=timepoints,
        index=idx,
    )

    return df


def assert_move_plan_equal(res: MovePlan, exp: MovePlan) -> None:
    """
    Assert that two [MovePlan][(p).db.rewriting.] are equal

    Parameters
    ----------
    res
        The result

    exp
        The expectation

    Raises
    ------
    AssertionError
        `res` and `exp` are not equal
    """
    # Check that the indexes are the same.
    # We convert to MultiIndex first as we don't care about the actual index values.
    pd.testing.assert_index_equal(  # type: ignore # pandas-stubs out of date
        pd.MultiIndex.from_frame(res.moved_index.reset_index()),
        pd.MultiIndex.from_frame(exp.moved_index.reset_index()),
        check_order=False,
    )
    pd.testing.assert_series_equal(
        res.moved_file_map, exp.moved_file_map, check_like=True
    )

    if res.rewrite_actions is None:
        assert exp.rewrite_actions is None
    else:
        if exp.rewrite_actions is None:
            msg = f"{exp.rewrite_actions=} while {res.rewrite_actions=}"
            raise AssertionError(msg)

        assert len(res.rewrite_actions) == len(exp.rewrite_actions)
        for res_rwa in res.rewrite_actions:
            for exp_rwa in exp.rewrite_actions:
                if res_rwa.from_file == exp_rwa.from_file:
                    break
            else:
                msg = f"Did not find pair for\n{res_rwa=}\nin\n{exp.rewrite_actions=}"
                raise AssertionError(msg)

            pd.testing.assert_index_equal(  # type: ignore # pandas-stubs out of date
                res_rwa.locator, exp_rwa.locator, check_order=False
            )
            assert res_rwa.to_file == exp_rwa.to_file

    if res.delete_paths is None:
        assert exp.delete_paths is None
    else:
        if exp.delete_paths is None:
            msg = f"{exp.delete_paths=} while {res.delete_paths=}"
            raise AssertionError(msg)

        assert set(res.delete_paths) == set(exp.delete_paths)


def changer(inv: float, factor: float, *, exponent: float = 1.0) -> float:
    """
    Change a value

    This is just meant as a helper for our tests
    """
    return cast(float, (inv * 2.0) ** exponent)



"""
Interfaces used throughout the db (database) module
"""

from __future__ import annotations

from pathlib import Path
from typing import Protocol, runtime_checkable

import pandas as pd


@runtime_checkable
class OpenSCMDBIndexBackend(Protocol):
    """
    Backend for (de-)serialising the index (and file map)

    Designed to be used with [OpenSCMDB][(m)]
    """

    ext: str
    """
    Extension to use with index files saved by this backend.
    """

    preserves_index: bool
    """
    Whether this backend preserves the `pd.MultiIndex` upon (de-)serialisation
    """

    @staticmethod
    def load_file_map(file_map_file: Path) -> pd.DataFrame:
        """
        Load the file map

        This is a low-level method
        that just handles the specifics of loading the index from disk.
        Working out the path from which to load the file map
        should happen in higher-level functions.

        Parameters
        ----------
        file_map_file
            File from which to load the file map

        Returns
        -------
        :
            Loaded file map

        Notes
        -----
        This returns a [pd.DataFrame][pandas.DataFrame].
        It is up to the user to cast this to a [pd.Series][pandas.Series]
        if they wish.
        """

    @staticmethod
    def load_index(index_file: Path) -> pd.DataFrame:
        """
        Load the index

        This is a low-level method
        that just handles the specifics of loading the index from disk.
        Working out the path from which to load the index
        should happen in higher-level functions.

        Parameters
        ----------
        index_file
            File from which to load the index

        Returns
        -------
        :
            Loaded index

        Notes
        -----
        This just loads the index directly from disk.
        If the index had a `pd.MultiIndex` when it was saved,
        this may or not be restored.
        It is up to the user
        to decide whether to do any `pd.MultiIndex` restoration or not,
        based on their use case and the value of `self.preserves_index`.
        We do not make this choice as converting back to a
        `pd.MultiIndex` can be a very expensive operation,
        and we want to give the user control over any such optimisations.
        """

    def save_file_map(
        self,
        file_map: pd.Series[Path],  # type: ignore # pandas confused about what it supports
        file_map_file: Path,
    ) -> None:
        """
        Save the file map to disk

        This is a low-level method
        that just handles the specifics of serialising the file map to disk.
        Working out what to save and in what path
        should happen in higher-level functions.

        Parameters
        ----------
        file_map
            File map to save

        file_map_file
            File in which to save the file map
        """

    def save_index(
        self,
        index: pd.DataFrame,
        index_file: Path,
    ) -> None:
        """
        Save the index to disk

        This is a low-level method
        that just handles the specifics of serialising the index to disk.
        Working out what to save and in what path
        should happen in higher-level functions.

        Parameters
        ----------
        index
            Index to save

        index_file
            File in which to save the index
        """


@runtime_checkable
class OpenSCMDBDataBackend(Protocol):
    """
    Backend for (de-)serialising data

    Designed to be used with [OpenSCMDB][(m)]
    """

    ext: str
    """
    Extension to use with data files saved by this backend.
    """

    preserves_index: bool
    """
    Whether this backend preserves the index of data upon (de-)serialisation
    """

    @staticmethod
    def load_data(data_file: Path) -> pd.DataFrame:
        """
        Load a data file

        This is a low-level method
        that just handles the specifics of loading the data from disk.
        Working out the path from which to load the data
        should happen in higher-level functions.

        Parameters
        ----------
        data_file
            File from which to load the data

        Returns
        -------
        :
            Loaded data

        Notes
        -----
        This just loads the data directly from disk.
        If the data had a `pd.MultiIndex` when it was saved,
        this may or not be restored.
        It is up to the user
        to decide whether to do any `pd.MultiIndex` restoration or not,
        based on their use case and the value of `self.preserves_index`.
        We do not make this choice as converting back to a
        `pd.MultiIndex` can be a very expensive operation,
        and we want to give the user control over any such optimisations.
        """

    @staticmethod
    def save_data(
        data: pd.DataFrame,
        data_file: Path,
    ) -> None:
        """
        Save data to disk

        This is a low-level method
        that just handles the specifics of serialising the data to disk.
        Working out what to save and in what path
        should happen in higher-level functions.

        Parameters
        ----------
        data
            Data to save

        data_file
            File in which to save the data
        """



"""
Manipulation of the index of data
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    P = TypeVar("P", pd.DataFrame, pd.Series[Any])

    import pandas.core.indexes.frozen


def convert_index_to_category_index(pandas_obj: P) -> P:
    """
    Convert the index's values to categories

    This can save a lot of memory and improve the speed of processing.
    However, it comes with some pitfalls.
    For a nice discussion of some of them,
    see [this article](https://towardsdatascience.com/staying-sane-while-adopting-pandas-categorical-datatypes-78dbd19dcd8a/).

    Parameters
    ----------
    pandas_obj
        Object whose index we want to change to categorical.

    Returns
    -------
    :
        A new object with the same data as `pandas_obj`
        but a category type index.
    """
    new_index = pd.MultiIndex.from_frame(
        pandas_obj.index.to_frame(index=False).astype("category")
    )

    if hasattr(pandas_obj, "columns"):
        return type(pandas_obj)(  # type: ignore # confusing mypy here
            pandas_obj.values,
            index=new_index,
            columns=pandas_obj.columns,
        )

    return type(pandas_obj)(
        pandas_obj.values,
        index=new_index,
    )


def update_index_from_candidates(
    indf: pd.DataFrame, candidates: pandas.core.indexes.frozen.FrozenList
) -> pd.DataFrame:
    """
    Update the index of data to align with the candidate columns as much as possible

    Parameters
    ----------
    indf
        Data of which to update the index

    candidates
        Candidate columns to use to create the updated index

    Returns
    -------
    :
        `indf` with its updated index.

        All columns of `indf` that are in `candidates`
        are used to create the index of the result.

    Notes
    -----
    This overwrites any existing index of `indf`
    so you will only want to use this function
    when you're sure that there isn't anything of interest
    already in the index of `indf`.
    """
    set_to_index = [v for v in candidates if v in indf.columns]
    res = indf.set_index(set_to_index)

    return res


def unify_index_levels(
    left: pd.MultiIndex, right: pd.MultiIndex
) -> tuple[pd.MultiIndex, pd.MultiIndex]:
    """
    Unify the levels on two indexes

    The levels are unified by simply adding NaN to any level in either `left` or `right`
    that is not in the level of the other index.

    This is differnt to [pd.DataFrame.align][pandas.DataFrame.align].
    [pd.DataFrame.align][pandas.DataFrame.align]
    will fill missing values with values from the other index if it can.
    We don't want that here.
    We want any non-aligned levels to be filled with NaN.

    The implementation also allows this to be performed on indexes directly
    (avoiding casting to a DataFrame
    and avoiding paying the price of aligning everything else
    or creating a bunch of NaN that we just drop straight away).

    The indexes are returned with the levels from `left` first,
    then the levels from `right`.

    Parameters
    ----------
    left
        First index to unify

    right
        Second index to unify

    Returns
    -------
    left_aligned :
        Left after alignment

    right_aligned :
        Right after alignment

    Examples
    --------
    >>> import pandas as pd
    >>>
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["a", "b"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['a', 'b', 'c'])

    >>> # Also fine if b has swapped levels
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["b", "a"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([( 8,  7, nan),
                (11, 10, nan)],
               names=['a', 'b', 'c'])

    >>> # Also works if a is 'inside' b
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["a", "b"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> unified_a
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['a', 'b', 'c'])

    >>> unified_b
    MultiIndex([(1, 2, 3),
                (4, 5, 6)],
               names=['a', 'b', 'c'])

    >>> # But, be a bit careful, this is now sensitive to a's column order
    >>> idx_a = pd.MultiIndex.from_tuples(
    ...     [
    ...         (7, 8),
    ...         (10, 11),
    ...     ],
    ...     names=["b", "a"],
    ... )
    >>> idx_b = pd.MultiIndex.from_tuples(
    ...     [
    ...         (1, 2, 3),
    ...         (4, 5, 6),
    ...     ],
    ...     names=["a", "b", "c"],
    ... )
    >>> unified_a, unified_b = unify_index_levels(idx_a, idx_b)
    >>> # Note that the names are `['b', 'a', 'c']` in the output
    >>> unified_a
    MultiIndex([( 7,  8, nan),
                (10, 11, nan)],
               names=['b', 'a', 'c'])

    >>> unified_b
    MultiIndex([(2, 1, 3),
                (5, 4, 6)],
               names=['b', 'a', 'c'])
    """
    if left.names == right.names:
        return left, right

    if (not left.names.difference(right.names)) and (  # type: ignore # pandas-stubs confused
        not right.names.difference(left.names)  # type: ignore # pandas-stubs confused
    ):
        return left, right.reorder_levels(left.names)  # type: ignore # pandas-stubs missing reorder_levels

    out_names = [*left.names, *[v for v in right.names if v not in left.names]]
    out_names_s = set(out_names)
    left_to_add = out_names_s.difference(left.names)
    right_to_add = out_names_s.difference(right.names)

    left_unified = pd.MultiIndex(  # type: ignore # pandas-stubs missing reorder_levels
        levels=[
            *left.levels,
            *[np.array([], dtype=right.get_level_values(c).dtype) for c in left_to_add],  # type: ignore # pandas-stubs confused
        ],
        codes=[
            *left.codes,
            *([np.full(left.shape[0], -1)] * len(left_to_add)),
        ],
        names=[
            *left.names,
            *left_to_add,
        ],
    ).reorder_levels(out_names)

    right_unified = pd.MultiIndex(  # type: ignore # pandas-stubs missing reorder_levels
        levels=[
            *[np.array([], dtype=left.get_level_values(c).dtype) for c in right_to_add],  # type: ignore # pandas-stubs confused
            *right.levels,
        ],
        codes=[
            *([np.full(right.shape[0], -1)] * len(right_to_add)),
            *right.codes,
        ],
        names=[
            *right_to_add,
            *right.names,
        ],
    ).reorder_levels(out_names)

    return left_unified, right_unified


def unify_index_levels_check_index_types(
    left: pd.Index[Any], right: pd.Index[Any]
) -> tuple[pd.MultiIndex, pd.MultiIndex]:
    """
    Unify the levels on two indexes

    This is just a thin wrapper around [unify_index_levels][(m).]
    that checks the the inputs are both [pd.MultiIndex][pandas.MultiIndex]
    before unifying the indices.

    Parameters
    ----------
    left
        First index to unify

    right
        Second index to unify

    Returns
    -------
    left_aligned :
        Left after alignment

    right_aligned :
        Right after alignment
    """
    if not isinstance(left, pd.MultiIndex):
        raise TypeError(left)

    if not isinstance(right, pd.MultiIndex):
        raise TypeError(right)

    return unify_index_levels(left, right)
